/* Copyright (c) 1992-1993  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Mac menu handling. */

#include "conq.h"
#include "mac.h"

#define side_may_select(unit) ((unit) && ((unit)->side == dside || dside->designer))

#define MAXWINDOWS 100

Map *create_map();

extern int gamestatesafe;
extern int interfacestatesafe;

extern int nummaps;

int autoselect = FALSE;

char *cursavename = NULL;

MenuHandle mapviewmenu = nil;
MenuHandle listviewmenu = nil;
MenuHandle closeupviewmenu = nil;

MenuHandle utypemenu = nil;
MenuHandle ttypemenu = nil;
MenuHandle mtypemenu = nil;
MenuHandle sidemenu = nil;
MenuHandle aitypemenu = nil;

WindowPtr *winmenuwins;

WindowPtr worldmapwin = nil;

/* Junk associated with debug output. */

#ifdef USE_CONSOLE
#ifdef THINK_C
#include <console.h>
#endif
#endif

FILE *ffp = NULL;

/* Set up all the menus, at least in their basic form. */

init_menus()
{
	int i;
	Handle menubar;
	MenuHandle menu;

	menubar = GetNewMBar(mbMain);
	SetMenuBar(menubar);
	/* Add the DAs etc as usual. */
	if ((menu = GetMHandle(mApple)) != nil) {
		AddResMenu(menu, 'DRVR');
	}
	/* Get the different versions of the view menu. */
	mapviewmenu = GetMenu(mViewMap);
	listviewmenu = GetMenu(mViewList);
	closeupviewmenu = GetMenu(mViewCloseup);
	/* Init submenus and popups. */
	if ((menu       = GetMenu(mViewFontSizes)) != nil)  InsertMenu(menu, -1);
	if ((menu       = GetMenu(mMagnifications)) != nil) InsertMenu(menu, -1);
	if ((menu       = GetMenu(mViewMapWeather)) != nil) InsertMenu(menu, -1);
	if ((sidemenu   = GetMenu(mSides)) != nil)          InsertMenu(sidemenu, -1);
	if ((utypemenu  = GetMenu(mUnitTypes)) != nil)      InsertMenu(utypemenu, -1);
	if ((mtypemenu  = GetMenu(mMaterialTypes)) != nil)  InsertMenu(mtypemenu, -1);
	if ((ttypemenu  = GetMenu(mTerrainTypes)) != nil)   InsertMenu(ttypemenu, -1);
	if ((aitypemenu = GetMenu(mAITypes)) != nil)        InsertMenu(aitypemenu, -1);
	/* Init the support for the Windows menu. */
	winmenuwins = (WindowPtr *) xmalloc(MAXWINDOWS * sizeof(WindowPtr));
	for (i = 0; i < MAXWINDOWS; ++i) {
		winmenuwins[i] = nil;
	}
	/* Done fiddling with menus, draw them. */
	DrawMenuBar();
}

/* Add a menu item that can be used to go to the given window.  Set the window title
   here too, ensures that the two match. */

add_window_menu_item(name, win)
char *name;
WindowPtr win;
{
	int i;
	MenuHandle menu;
	Str255 tmpstr;

	for (i = 0; i < 100; ++i) {
		if (winmenuwins[i] == nil) {
			c2p(name, tmpstr);
			SetWTitle(win, tmpstr);
			winmenuwins[i] = win;
			if ((menu = GetMHandle(mWindows)) != nil) {
				AppendMenu(menu, tmpstr);
				return;
			}
		}
	}
}

/* Remove the menu item that can be used to select a given window. */

remove_window_menu_item(win)
WindowPtr win;
{
	int i;
	MenuHandle menu;

	for (i = 0; i < 100; ++i) {
		if (winmenuwins[i] == win) {
			winmenuwins[i] = nil;
			/* (should release and reconstruct menu instead) */
			if ((menu = GetMHandle(mWindows)) != nil) {
				DisableItem(menu, i + miWindowsFirst);
			}
			return;
		}
	}
}

/* Fill up the side menu. */

build_side_menu()
{
	Side *side2;
	Str255 sidename;

	if (CountMItems(sidemenu) < 1 /* numsides + 1 */) {
		for_all_sides(side2) {
			shortest_side_title(side2, spbuf);
			sanitize_for_menu(spbuf);
			c2p(spbuf, sidename);
			AppendMenu(sidemenu, sidename);
			EnableItem(sidemenu, side_number(side2));
		}
		/* (only if indep units possible?) */
		AppendMenu(sidemenu, "\pindependent");
	}
}

build_unit_type_menu()
{
	int u;
	Str255 typename;

	if (utypemenu != nil && CountMItems(utypemenu) < 1) {
		for_all_unit_types(u) {
			sprintf(spbuf, "%s", u_type_name(u));
#if 0 /* when new routine becomes available */
			c2p(spbuf, typename);
			AppendMenu(utypemenu, "\px");
			SetMenuItemText(utypemenu, u + 1, typename);
#else
			sanitize_for_menu(spbuf);
			c2p(spbuf, typename);
			AppendMenu(utypemenu, typename);
#endif
			EnableItem(utypemenu, u + 1);
		}
	}
}

build_material_type_menu()
{
	int m;
	Str255 typename;

	if (mtypemenu != nil && CountMItems(mtypemenu) < 1 && nummtypes > 0) {
		for_all_material_types(m) {
			sprintf(spbuf, "%s", m_type_name(m));
			sanitize_for_menu(spbuf);
			c2p(spbuf, typename);
			AppendMenu(mtypemenu, typename);
			EnableItem(mtypemenu, m + 1);
		}
	}
}

build_terrain_type_menu()
{
	int t;
	Str255 typename;

	if (ttypemenu != nil && CountMItems(ttypemenu) < 1) {
		for_all_terrain_types(t) {
			sprintf(spbuf, "%s", t_type_name(t));
			sanitize_for_menu(spbuf);
			c2p(spbuf, typename);
			AppendMenu(ttypemenu, typename);
			EnableItem(ttypemenu, t);
		}
	}
}

build_ai_type_menu()
{
	if (aitypemenu != nil && CountMItems(aitypemenu) < 1) {
		AppendMenu(aitypemenu, "\pnone");
		EnableItem(aitypemenu, 1);
		AppendMenu(aitypemenu, "\pmplayer");
		EnableItem(aitypemenu, 2);
	}
}

/* This removes chars that are specially recognized by the Menu Manager. */

sanitize_for_menu(buf)
char *buf;
{
	int i;

	/* Replace a leading hypen with an underscore. */
	if (buf[0] == '-') {
		buf[0] = '_';
	}
	/* Replace special chars with blanks. */
	for (i = 0; buf[i] != '\0'; ++i) {
		switch (buf[i]) {
			case ';':
			case '!':
			case '<':
			case '/':
			case '(':
				buf[i] = ' ';
				break;
			default:
				break;
		}
	}
}

/* Decipher and do a menu command. */

do_menu_command(which)
long which;
{
	short menuid, menuitem, itemHit, daRefNum, ditem, i;
	Str255 daname;
	Boolean handledbyda;

	menuid = HiWord(which);
	menuitem = LoWord(which);
	DGprintf("menu %d, item %d\n", menuid, menuitem);
	switch (menuid) {
		case mApple:
			switch (menuitem) {
				case miAppleAbout:
					do_about_box();
					break;
				case miAppleHelp:
					help_dialog();
					break;
#ifdef DEBUGGING
				/* (should eventually make all this less visible, like on keys) */
				case 3:
					Debug = !Debug;
					do_debugging(Debug);
					break;
				case 4:
					DebugG = !DebugG;
					do_debugging(DebugG);
					break;
				case 5:
					DebugM = !DebugM;
					do_debugging(DebugM);
					break;
#endif /* DEBUGGING */
				default:
					GetItem(GetMHandle(mApple), menuitem, daname);
					daRefNum = OpenDeskAcc(daname);
			}
			break;
		case mFile:
			switch (menuitem) {
				case miFileNew:
					/* (should reset state of everything first) */
					new_game_dialog();
					break;
				case miFileOpen:
					/* (should reset state of everything first) */
					open_game_dialog();
					break;
				case miFileConnect:
					connect_game_dialog();
					break;
				case miFileSave:
					save_the_game(FALSE);
					break;
				case miFileSaveAs:	
					save_the_game(TRUE);
					break;
				case miFilePreferences:
					set_preferences();
					break;
				case miFilePrint:
					/* (should print the front window reasonably) */
					break;
				case miFileQuit:
					quit_the_game();
					break;
			}
			break;
		case mEdit:
			/* handledbyda = SystemEdit(menuitem-1); */
			switch (menuitem)  {
				case miEditCut:
					/* disband_selection(); */
					/* (should depend on whether this is a map or a text field) */
					break;
				case miEditSelectAll:
					do_select_all_mi();
					break;
#ifdef DESIGNERS
				case miEditDesign:
					if (!dside->designer) {
						enable_designing(FALSE);
					} else {
						disable_designing();
					}
					break;
#endif
			}
			break;
		case mFind:
			switch (menuitem) {
				case miFindPrevious:
					do_find_previous_mi();
					break;
				case miFindNext:
					do_find_next_mi();
					break;
				case miFindLocation:
					do_find_location_mi();
					break;
				case miFindUnitByName:
					do_find_unit_by_name_mi();
					break;
				case miFindSelected:
					do_find_selected_mi();
					break;
			}
			break;
		case mPlay:
			switch (menuitem) {
				case miPlayCloseup:
					do_closeup_mi();
					break;
				case miPlayMove:
					do_move_mi();
					break;
				case miPlayPatrol:
					do_patrol_mi();
					break;
				case miPlayReturn:
					do_return_mi();
					break;
				case miPlayWake:
					do_wake_mi();
					break;
				case miPlaySleep:
					do_sleep_mi();
					break;
				case miPlayDoneMoving:
					do_done_moving_mi();
					break;
				case miPlayBuild:
					do_construction_mi();
					break;
				case miPlayRepair:
					do_repair_mi();
					break;
				case miPlayAttack:
					do_attack_mi();
					break;
				case miPlayOverrun:
					do_overrun_mi();
					break;
				case miPlayFire:
					do_fire_mi();
					break;
				case miPlayDetonate:
					do_detonate_mi();
					break;
				case miPlayTake:
					break;
				case miPlayDrop:
					break;
				case miPlayGive:
					/* has submenu mSides */
					break;
				case miPlayDisband:
					do_disband_mi();
					break;
				case miPlayRename:
					do_rename_mi();
					break;
			}
			break;
		case mSides:
			do_give_mi(menuitem);
			break;
		case mSide:
			switch (menuitem) {
				case miSideCloseup:
#if 0
					/* Eventually allow designers to look at other sides. */
					if (dside->ui->sidecloseups == NULL) {
						create_side_closeup(dside);
					}
					ShowWindow(dside->ui->sidecloseups->dialog);
#endif
					break;
				case miSideFinishedTurn:
					finish_turn(dside);
					run_game(10);
					break;
				case miSideMoveOnClick:
					moveonclick = !moveonclick;
					/* (should force updates of map windows to reflect?) */
					break;
				case miSideAutoSelect:
					/* Toggle auto-selection of units. */
					autoselect = !autoselect;
					break;
				case miSideAutoFinish:
					/* Toggle auto-selection of units. */
					dside->autofinish = !dside->autofinish;
					break;
				case miSideRename:
					side_rename_dialog(dside);
					break;
			}
			break;
		case mWindows:
			switch (menuitem) {
				case miWindowsGame:
					if (gamewin == nil) {
						create_game_window();
					}
					ShowWindow(gamewin);
					SelectWindow(gamewin);
					break;
				case miWindowsHistory:
					if (historywin == nil) {
						create_history_window();
					}
					ShowWindow(historywin);
					SelectWindow(historywin);
					break;
				/* agreements list etc */
				case miWindowsNewList:
					create_list();
					break;
				case miWindowsNewMap:
					create_map(5 /* should be a pref */);
					break;
				case miWindowsWorldMap:
					if (worldmapwin == nil) {
						int dim = min(200 / area.width, 200 / area.height);
						int power;

						for (power = 0; power < NUMPOWERS; ++power) {
							if (hws[power] <= dim && hws[power+1] > dim) break;
						}
						worldmapwin = (create_map(power))->window;
					}
					ShowWindow(worldmapwin);
					SelectWindow(worldmapwin);
					break;
				default:
					{
						WindowPtr win = winmenuwins[menuitem - miWindowsFirst];
						
						if (win != nil) {
							SelectWindow(win);
						}
					}
					break;
			}
			break;
		case mViewMap:
			switch (menuitem)  {
				case miViewCloser:
					zoom_map(map_from_window(FrontWindow()), 1);
					break;
				case miViewFarther:
					zoom_map(map_from_window(FrontWindow()), -1);
					break;
				case miViewNames:
					toggle_map_names(map_from_window(FrontWindow()));
					break;
				case miViewPeople:
					toggle_map_people(map_from_window(FrontWindow()));
					break;
				case miViewElevations:
					toggle_map_elevations(map_from_window(FrontWindow()));
					break;
				case miViewDaylight:
					toggle_map_lighting(map_from_window(FrontWindow()));
					break;
				case miViewGrid:
					toggle_map_grid(map_from_window(FrontWindow()));
					break;
				case miViewTopline:
					toggle_map_topline(map_from_window(FrontWindow()));
					break;
				case miViewOtherMaps:
					toggle_map_other_maps(map_from_window(FrontWindow()));
					break;
			}
			break;
		case mViewWeather:
			switch (menuitem)  {
				case 1:
					break;
				case 2:
					break;
				case 3:
					break;
				case 4:
					break;
			}
		case mViewList:
			switch (menuitem)  {
				case miViewIconSize:
					toggle_list_large_icons(list_from_window(FrontWindow()));
					break;
			}
			break;
		case mAITypes:
			switch (menuitem) {
				case 1:
					if (side_has_ai(dside)) {
						dside->ai = NULL;
						dside->player->aitypename = NULL;
						/* should clear all unit plans etc? */
					}
					break;
				case 2:
					if (!side_has_ai(dside)) {
						dside->player->aitypename = "mplayer";
						init_ai(dside);
					}
					break;
			}
			break;
		case mMagnifications:
			set_map_mag(map_from_window(FrontWindow()), menuitem - 1);
			break;
		case mMaterialTypes:
			toggle_map_materials(map_from_window(FrontWindow()), menuitem - 1);
			break;
	}
	HiliteMenu(0);
}

/* Display the "About..." box. */

/* This dialog does double duty - initially it chooses New/Open/Connect/Quit,
   later on it's just a normal About box. */

/* (should make initial 4-way choice more stylish/interesting) */

do_about_box()
{
	short ditem;
	Str255 tmpstr;
	WindowPtr win;
	PicHandle pic;
	short itemtype;  Handle itemhandle;  Rect itemrect;

	win = GetNewDialog(dAbout, NULL, (DialogPtr) -1L);
	GetDItem(win, diAboutVersion, &itemtype, &itemhandle, &itemrect);
	c2p(version_string(), tmpstr);
	SetIText(itemhandle, tmpstr);
	GetDItem(win, diAboutCopyright, &itemtype, &itemhandle, &itemrect);
	c2p(copyright_string(), tmpstr);
	SetIText(itemhandle, tmpstr);
	/* (should decide about b/w vs color splash pict here) */
	if (hasColorQD) {
		GetDItem(win, diAboutSplash, &itemtype, &itemhandle, &itemrect);
		pic = (PicHandle) GetResource('PICT', pSplashColor);
		SetDItem(win, diAboutSplash, itemtype, (Handle) pic, &itemrect);
	}
	/* What this box looks like depends on whether we're in a game. */
	if (ingame) {
		/* Make this look like a "normal" About box by hiding all the buttons. */
		GetDItem(win, diAboutNew, &itemtype, &itemhandle, &itemrect);
		HideControl((ControlHandle) itemhandle);
		GetDItem(win, diAboutOpen, &itemtype, &itemhandle, &itemrect);
/*		HideControl((ControlHandle) itemhandle); */
		SetCTitle((ControlHandle) itemhandle, "\pOK");
		GetDItem(win, diAboutConnect, &itemtype, &itemhandle, &itemrect);
		HideControl((ControlHandle) itemhandle);
		GetDItem(win, diAboutQuit, &itemtype, &itemhandle, &itemrect);
		HideControl((ControlHandle) itemhandle);
		ShowWindow(win);
		SelectWindow(win);
/*		while (!Button()); */
		ModalDialog(NULL, &ditem);
		HideWindow(win);
	} else {
		/* (should enable/disable connect button) */
		if (can_connect()) {
		} else {
		}
		ShowWindow(win);
		draw_default_button(win, diAboutNew);
		ModalDialog(NULL, &ditem);
		/* We don't mess around with fanciness here, just return the ditem and let
		   caller decide what to do, if anything. */
	}
	DisposDialog(win);
	return ditem;
}


#ifdef DEBUGGING

do_debugging(flag)
int flag;
{
	/* Debug output uses a file. */
	if (flag) {
		if (ffp != NULL) {
			fclose(ffp);
			ffp = NULL;
		}
		if (ffp == NULL) {
			ffp = fopen("Xconq.DebugOut", "a");
		}
		if (ffp != NULL) {
			if (Debug) dfp = ffp;
			if (DebugG) dgfp = ffp;
			if (DebugM) dmfp = ffp;
		}
	} else {
		if (!(Debug || DebugG || DebugM)) {
			if (ffp != NULL) {
				fclose(ffp);
				ffp = NULL;
			}
		}
	}
}

#endif /* DEBUGGING */

/* This routine does both "save" and "save as". */

save_the_game(askname)
int askname;
{
	extern CursHandle watchcursor;

#ifdef DESIGNERS
	if (dside->designer) {
		designer_save_dialog();
	} else
#endif
	if (askname || cursavename == NULL) {
		if (cursavename == NULL) cursavename = savefile_name();
		/* should bring up dialog to choose name */
		/* Do the save proper. */
		SetCursor(*watchcursor);
		write_entire_game_state(cursavename);
		setfiletype(cursavename);
	} else {
		/* Just save to the same file as before. */
		SetCursor(*watchcursor);
		write_entire_game_state(cursavename);
		setfiletype(cursavename);
	}
/*	adjust_cursor(?, ?); */
}

set_preferences()
{
	short done = FALSE, ditem;
	DialogPtr win;
	short itemtype;  Handle itemhandle;  Rect itemrect;
	extern int checkpointinterval;

	/* (should have gotten prefs from file already) */
	win = GetNewDialog(dPreferences, NULL, (DialogPtr) -1L);
	/* Set the current preferences into the items. */
	GetDItem(win, diPrefsGrid, &itemtype, &itemhandle, &itemrect);
	SetCtlValue((ControlHandle) itemhandle, defaultdrawgrid);
	GetDItem(win, diPrefsNames, &itemtype, &itemhandle, &itemrect);
	SetCtlValue((ControlHandle) itemhandle, defaultdrawnames);
	GetDItem(win, diPrefsCheckpoint, &itemtype, &itemhandle, &itemrect);
	SetCtlValue((ControlHandle) itemhandle, (checkpointinterval > 0));
	ShowWindow(win);
	while (!done) {
		ModalDialog(NULL, &ditem);
		switch (ditem) {
			case diPrefsOK:
				/* Actually change the program's variables. */
				GetDItem(win, diPrefsGrid, &itemtype, &itemhandle, &itemrect);
				defaultdrawgrid = GetCtlValue((ControlHandle) itemhandle);
				GetDItem(win, diPrefsNames, &itemtype, &itemhandle, &itemrect);
				defaultdrawnames = GetCtlValue((ControlHandle) itemhandle);
				GetDItem(win, diPrefsCheckpoint, &itemtype, &itemhandle, &itemrect);
				checkpointinterval = (GetCtlValue((ControlHandle) itemhandle) ? 5 : 0);
				/* Fall into next case. */
			case diPrefsCancel:
				done = TRUE;
				break;
			case diPrefsGrid:
			case diPrefsNames:
				/* Toggle check boxes. */
				GetDItem(win, ditem, &itemtype, &itemhandle, &itemrect);
				SetCtlValue((ControlHandle) itemhandle, !GetCtlValue((ControlHandle) itemhandle));
				break;
			case diPrefsCheckpoint:
				GetDItem(win, ditem, &itemtype, &itemhandle, &itemrect);
				SetCtlValue((ControlHandle) itemhandle, !GetCtlValue((ControlHandle) itemhandle));
				/* Should enable/disable interval-setting ditems. */
				break;
			case diPrefsInterval:
				break;
		}
	}
	DisposDialog(win);
}

/* Attempt to quit. */

/* (should warn if quitting is tantamount to resigning) */
/* (should close any open windows left) */

quit_the_game()
{
	extern int gamestatesafe, interfacestatesafe;
	extern int endofgame;
	Boolean closed;
	WindowPtr win;

	if (endofgame) {
		/* Game is over, nothing to save. */
		ExitToShell();
	} else if (!gamestatesafe || !interfacestatesafe) {
		switch (CautionAlert(aQuitGame, nil)) {
			case aiQuitGameOK:
				/* (should maybe do game resignation, update other sides) */
				SetCursor(*watchcursor);
				save_the_game(FALSE);
				SetCursor(&qd.arrow);
				ExitToShell();
				break;
			case aiQuitGameCancel:
				break;
			case aiQuitGameDontSave:
#if 0
	switch (CautionAlert(aResignGame, nil)) {
		case aiResignOK:
			/* (should do game resignation, update other sides) */
			break;
		case aiResignCancel:
			break;
	}
#endif
				ExitToShell();
				break;
		}
	} else {
		/* (should still confirm that we're not avoiding our fate...) */
		ExitToShell();
	}
}

do_select_all_mi()
{
	int i;
	Map *map;
	Unit *unit;
	Side *side;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for_all_units(side, unit) {
			if (side_may_select(unit)) {
				select_unit_on_map(map, unit);
			}
		}
		/* Do all the drawing at once. */
		draw_selections(map);
	}
}

do_find_previous_mi()
{
	Map *map;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		select_previous_unit(map);
	}
}

do_find_next_mi()
{
	Map *map;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		select_next_unit(map);
	}
}

/* (should ask for x,y coords, focus/center on that position) */

do_find_location_mi()
{
	Map *map;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
	}
}

/* (should ask for a textual name/select from a gazetteer, scroll to that unit) */

do_find_unit_by_name_mi()
{
	Map *map;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
	}
}

do_find_selected_mi()
{
	Map *map;
	List *list;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
	} else if ((list = (List *) list_from_window(FrontWindow())) != NULL) {
		scroll_to_selected_unit_in_list(list);
	} else {
		/* nothing to do */
	}
}

/* Menu items/commands for unit play action. */


do_closeup_mi()
{
	int i;
	Map *map;
	List *list;
	Unit *unit;
	UnitCloseup *unitcloseup;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				if ((unitcloseup = find_unit_closeup(unit)) != NULL) {
					ShowWindow(unitcloseup->window);
					SelectWindow(unitcloseup->window);
				} else {
					create_unit_closeup(unit);
				}
			}
		}
	} else if ((list = (List *) list_from_window(FrontWindow())) != NULL) {
	}
}

do_move_mi()
{
	SysBeep(20);
}

do_patrol_mi()
{
	int i;
	Map *map;
	Unit *unit;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				if (can_move(unit)) {
				}
			}
		}
	}
}

do_return_mi()
{
	int i;
	Map *map;
	Unit *unit;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				if (can_move(unit)) {
					/* call a generic return planner */
				}
			}
		}
	}
}

do_done_moving_mi()
{
	do_reserve_command(-1);
}

do_reserve_command(value)
int value;
{
	int i;
	Map *map;
	Unit *unit;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				if (unit->plan) {
					unit->plan->reserve = (value >= 0 ? value : !unit->plan->reserve);
				}
			}
		}
		/* (should eventually be unneeded because unit mods cause correct updates) */
		update_side_display(dside, dside, TRUE);
	}
}

do_wake_mi()
{
	do_sleep_command(0);
}

do_sleep_mi()
{
	do_sleep_command(1);
}

do_sleep_command(value)
int value;
{
	int i;
	Map *map;
	Unit *unit;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				if (unit->plan) {
					unit->plan->asleep = value;
				}
			}
		}
		/* (should eventually be unneeded because unit mods cause correct updates) */
		update_side_display(dside, dside, TRUE);
	}
}

do_construction_mi()
{
	int i;
	Map *map;
	Unit *unit;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				if (can_build(unit)) {  /* and completers and researchers too? */
					enable_construction();
					select_unit_in_construction_window(unit);
					return;
				}
			}
		}
	}
}

/* Create and/or bring up the construction planning window. */

enable_construction()
{
	if (buildwin == nil) {
		create_construction_window();
	}
	if (buildwin != nil) {
		ShowWindow(buildwin);
		SelectWindow(buildwin);
	}
}

do_repair_mi()
{
	int i;
	Map *map;
	Unit *unit;

	/* (make a "apply_to_all_selected" function...) */	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				if (0 /*can_move(unit) */) {
					/* set up a reasonable repair action or task */
				}
			}
		}
	}
}

/* (should find one unit to attack, do modal dialog if != 1) */

do_attack_mi()
{
	int i;
	Map *map;
	Unit *atker, *other;
	Side *side;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((atker = map->selections[i]) != NULL) {
				for_all_units(side, other) {
					if (other != atker
						&& other->side != atker->side
						&& distance(atker->x, atker->y, other->x, other->y) == 1) {
							prep_attack_action(atker, other, 100);
							break;
					}
				}
			}
		}
	}
}

do_attack_command()
{
	int i, x, y, numcould = 0, numnot = 0;
	Point target;
	Map *map;
	Unit *atker, *other;
	Side *side;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		GetMouse(&target);
		nearest_unit(map, target.h, target.v, &other);
		for (i = 0; i < map->numselections; ++i) {
			if ((atker = map->selections[i]) != NULL) {
				if (other != NULL
					&& other != atker
					&& other->side != atker->side) {
					if (distance(atker->x, atker->y, other->x, other->y) <= 1) {
						prep_attack_action(atker, other, 100);
					} else {
						prep_fire_at_action(atker, other, 0);
					}
					++numcould;
				} else {
					++numnot;
				}
			}
		}
	}
	/* If everybody that was asked to do the action couldn't, beep once. */
	if (numcould == 0 && numnot > 0) SysBeep(20);
}

do_overrun_mi()
{
	SysBeep(20);
}

do_fire_mi()
{
	SysBeep(20);
}

/* (should decide how to detonate based on selections and modal input) */

do_detonate_mi()
{
	int i;
	Map *map;
	Unit *unit;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				prep_detonate_action(unit, unit->x, unit->y, unit->z, unit);
			}
		}
	}
}

/* This command samples the current mouse to get the desired attack location. */

do_detonate_command()
{
	int i, x, y;
	Point detpoint;
	Map *map;
	Unit *unit;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		GetMouse(&detpoint);
		nearest_cell(map, detpoint.h, detpoint.v, &x, &y);
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				prep_detonate_action(unit, x, y, unit->z, unit);
			}
		}
	}
}

do_give_mi(mi)
int mi;
{
	int i;
	Side *side = side_n(mi); /* note that side_n returns NULL for mi == numsides */
	Map *map;
	Unit *unit;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
#ifdef DESIGNERS
				if (dside->designer) {
					unit_changes_side(unit, side, -1, -1);
					update_unit_display(dside, unit, TRUE);
				} else
#endif /* DESIGNERS */
				{
					prep_change_side_action(unit, side, unit);
				}
			}
		}
	}
}

do_disband_mi()
{
	int i;
	Map *map;
	Unit *unit;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
#ifdef DESIGNERS
				if (dside->designer) {
					kill_unit(unit, -1);
				} else
#endif /* DESIGNERS */
				if (!completed(unit)) {
					kill_unit(unit, DISBAND);
				} else if (unit->act && unit->act->acp > 0) {
					prep_disband_action(unit, unit);
				} else {
					/* try to find a nearby unit to do it */
				}
			}
		}
	}
}

/* This just renames the first of the selected units. */

do_rename_mi()
{
	int i;
	Map *map;
	List *list;
	Unit *unit;
	
	if ((map = (Map *) map_from_window(FrontWindow())) != NULL) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				unit_rename_dialog(unit);
				break;
			}
		}
	} else if ((list = (List *) list_from_window(FrontWindow())) != NULL) {
		if ((unit = (Unit *) selected_unit_in_list(list)) != NULL) {
			unit_rename_dialog(unit);
		}
	}
}


/* The state of menu items changes to reflect selected units, etc. */

adjust_menus()
{
	int i, m, nextmi, numwins;
	MenuHandle menu;
	Map *frontmap;
	List *frontlist;
	Unit *unit;

	frontmap = map_from_window(FrontWindow());
	frontlist = list_from_window(FrontWindow());
	/* Certain menus need to be built after the game is loaded. */
	if (ingame) {
		build_material_type_menu();
		build_side_menu();
		build_ai_type_menu();
	}
#if 0
	if (is_da_window(FrontWindow())) {
		if ((menu = GetMHandle(mEdit)) != nil) {
		}
		return;
	}
#endif
#ifdef DEBUGGING
	if ((menu = GetMHandle(mApple)) != nil) {
		CheckItem(menu, 3, Debug);
		CheckItem(menu, 4, DebugG);
		CheckItem(menu, 5, DebugM);
	}
#endif /* DEBUGGING */
	if ((menu = GetMHandle(mFile)) != nil) {
		/* Availability of file menu items depends on whether we're in a game. */
		if (ingame) {
			DisableItem(menu, miFileNew);
			DisableItem(menu, miFileOpen);
			DisableItem(menu, miFileConnect);
			EnableItem(menu, miFileSave);
			EnableItem(menu, miFileSaveAs);
		} else {
			if (numutypes == 0 /* should be !game_already_loaded? */) {
				EnableItem(menu, miFileNew);
				EnableItem(menu, miFileOpen);
			}
			if (can_connect()) {
				EnableItem(menu, miFileConnect);
			}
			DisableItem(menu, miFileSave);
			DisableItem(menu, miFileSaveAs);
		}
	}
	if ((menu = GetMHandle(mEdit)) != nil) {
		/* Edit menu is always available, but most items may be disabled. */
		/* (should do the other items also) */
		if (ingame) {
			EnableItem(menu, miEditSelectAll); /* just for maps and lists? */
#ifdef DESIGNERS
			CheckItem(menu, miEditDesign, dside->designer);
#else
			DisableItem(menu, miEditDesign);
#endif /* DESIGNERS */
		} else {
			DisableItem(menu, miEditSelectAll);
		}
	}
	if ((menu = GetMHandle(mPlay)) != nil) {
		if (ingame) {
			EnableItem(menu, 0);
			/* Disable everything first, then selectively re-enable. */
			DisableItem(menu, miPlayCloseup);
			DisableItem(menu, miPlayMove);
			DisableItem(menu, miPlayPatrol);
			DisableItem(menu, miPlayReturn);
			DisableItem(menu, miPlayWake);
			DisableItem(menu, miPlaySleep);
			DisableItem(menu, miPlayDoneMoving);
			DisableItem(menu, miPlayBuild);
			DisableItem(menu, miPlayRepair);
			DisableItem(menu, miPlayAttack);
			DisableItem(menu, miPlayOverrun);
			DisableItem(menu, miPlayFire);
			DisableItem(menu, miPlayDetonate);
			DisableItem(menu, miPlayTake);
			DisableItem(menu, miPlayDrop);
			DisableItem(menu, miPlayGive);
			DisableItem(menu, miPlayDisband);
			DisableItem(menu, miPlayRename);
			/* Note that command enabling is accumulative for all units. */
			if (frontmap != NULL) {
				for (i = 0; i < frontmap->numselections; ++i) {
					if ((unit = frontmap->selections[i]) != NULL) {
						enable_commands_for_unit(menu, unit);
					}
				}
			} else if (frontlist != NULL) {
				if ((unit = (Unit *) selected_unit_in_list(frontlist)) != NULL) {
					enable_commands_for_unit(menu, unit);
				}
			}
		} else {
			/* We're not even in the game yet, turn entire menu off. */
			DisableItem(menu, 0);
		}
	}
	if ((menu = GetMHandle(mSide)) != nil) {
		if (ingame) {
			EnableItem(menu, 0);
			CheckItem(menu, miSideMoveOnClick, moveonclick);
			CheckItem(menu, miSideAutoSelect, autoselect);
			CheckItem(menu, miSideAutoFinish, dside->autofinish);
			if (dside->nameslocked) {
				DisableItem(menu, miSideRename);
			} else {
				EnableItem(menu, miSideRename);
			}
		} else {
			/* We're not even in the game yet, turn entire menu off. */
			DisableItem(menu, 0);
		}
	}
	if ((menu = GetMHandle(mWindows)) != nil) {
		if (ingame) {
			EnableItem(menu, 0);
			/* Every item is always enabled. (should be anyway) */
			CheckItem(menu, miWindowsGame,
				(gamewin && ((WindowPeek) gamewin)->visible));
			CheckItem(menu, miWindowsHistory,
				(historywin && ((WindowPeek) historywin)->visible));
			numwins = CountMItems(menu) - miWindowsFirst;
			for (i = 0; i < numwins; ++i) {
				CheckItem(menu, i + miWindowsFirst, (winmenuwins[i] == FrontWindow()));
			}
		} else {
			/* We're not even in the game yet, turn entire menu off. */
			DisableItem(menu, 0);
		}
	}
	/* If a map window is frontmost, install and adjust the map view menu. */
	if (frontmap != NULL) {
		/* Delete the list view menu if that's what's installed currently. */
		/* (should be able to flush any non-map-view menu) */
		if ((menu = GetMHandle(mViewList)) != nil) {
			DeleteMenu(mViewList);
		}
		/* Make sure the map view menu is installed (is always at the end). */
		InsertMenu(mapviewmenu, 0);
		if ((menu = GetMHandle(mViewMap)) != nil) {
			EnableItem(menu, 0);
			if (frontmap->power == 0) {
				DisableItem(menu, miViewFarther);
			} else {
				EnableItem(menu, miViewFarther);
			}
			if (frontmap->power == NUMPOWERS-1) {
				DisableItem(menu, miViewCloser);
			} else {
				EnableItem(menu, miViewCloser);
			}
			EnableItem(menu, miViewGrid);
			CheckItem(menu, miViewGrid, frontmap->drawgrid);
			EnableItem(menu, miViewTopline);
			CheckItem(menu, miViewTopline, (frontmap->toph > 0));
			EnableItem(menu, miViewNames);
			CheckItem(menu, miViewNames, frontmap->drawnames);
			if (people_sides_defined()) {
				EnableItem(menu, miViewPeople);
				CheckItem(menu, miViewPeople, frontmap->drawpeople);
			} else {
				DisableItem(menu, miViewPeople);
			}
			if (elevations_defined()) {
				EnableItem(menu, miViewElevations);
				CheckItem(menu, miViewElevations, frontmap->drawelevations);
			} else {
				DisableItem(menu, miViewElevations);
			}
			if (nummtypes > 0) {
				EnableItem(menu, miViewMaterials);
				/* (should do checkmark if anything in submenu is on) */
			} else {
				DisableItem(menu, miViewMaterials);
			}
			if (world.daylength != 1) {
				EnableItem(menu, miViewDaylight);
				CheckItem(menu, miViewDaylight, frontmap->drawlighting);
			} else {
				DisableItem(menu, miViewDaylight);
			}
			if (nummaps > 1) {
				EnableItem(menu, miViewOtherMaps);
				CheckItem(menu, miViewOtherMaps, frontmap->drawothermaps);
			} else {
				DisableItem(menu, miViewOtherMaps);
			}
		}
	} else if (frontlist != NULL) {
		/* Do the list view menu similarly. */
		if ((menu = GetMHandle(mViewMap)) != nil) {
			DeleteMenu(mViewMap);
		}
		/* Make sure the menu is actually installed (is always at the end). */
		InsertMenu(listviewmenu, 0);
		if ((menu = GetMHandle(mViewList)) != nil) {
			EnableItem(menu, 0);
			DisableItem(menu, miViewByType);
			DisableItem(menu, miViewByName);
			EnableItem(menu, miViewBySide);
			CheckItem(menu, miViewBySide, 1);
			DisableItem(menu, miViewByActOrder);
			DisableItem(menu, miViewByLocation);
			DisableItem(menu, miViewWithTransport);
			DisableItem(menu, miViewWithCommander);
			EnableItem(menu, miViewIconSize);
			CheckItem(menu, miViewIconSize, frontlist->largeicons);
		}
	} else {
		/* For any other window, disable any or all of the view menus, as needed. */
		if ((menu = GetMHandle(mViewList)) != nil) {
			DisableItem(menu, 0);
		}
		if ((menu = GetMHandle(mViewMap)) != nil) {
			DisableItem(menu, 0);
		}
	}
	if ((menu = GetMHandle(mMagnifications)) != nil) {
		/* Always on. */
		EnableItem(menu, 0);
		if (frontmap != NULL) {
			for (i = 0; i < NUMPOWERS; ++i) {
				CheckItem(menu, i + 1, (i == frontmap->power));
			}
		}
	}
	if ((menu = GetMHandle(mMaterialTypes)) != nil) {
		/* Always on, if any material types defined. */
		EnableItem(menu, (nummtypes > 0));
		if (frontmap != NULL) {
			for_all_material_types(m) {
				CheckItem(menu, m + 1, frontmap->drawmaterials[m]);
			}
		}
	}
	if ((menu = GetMHandle(mAITypes)) != nil) {
		/* Always on. */
		EnableItem(menu, 0);
		CheckItem(menu, 1, (!side_has_ai(dside)));
		CheckItem(menu, 2, (side_has_ai(dside)));
	}
	/* Everything has been tweaked, redraw the menu bar. */
	DrawMenuBar();
}

/* Tweak the command menu according to what the unit can do. */
/* (some of these should have the menu text tweaked also, terms weird sometimes) */

enable_commands_for_unit(menu, unit)
MenuHandle menu;
Unit *unit;
{
	int u = unit->type, u2;

	/* If we can select the unit at all, we can always get a closeup of it. */
	EnableItem(menu, miPlayCloseup);
	/* (how to do checkmarks if multiple units selected?) */
	CheckItem(menu, miPlayCloseup, (find_unit_closeup(unit) != NULL));
	/* If we don't actually have any control over the unit, nothing more to do. */
	if (!side_controls_unit(dside, unit)) return;
	if (can_move_at_all(unit)) {
		EnableItem(menu, miPlayMove);
		EnableItem(menu, miPlayPatrol);
		/* also check for places to return to? */
		EnableItem(menu, miPlayReturn);
	}
	if (unit->plan) {
		EnableItem(menu, miPlayWake);
		CheckItem(menu, miPlayWake, !unit->plan->asleep);
		EnableItem(menu, miPlaySleep);
		CheckItem(menu, miPlaySleep, unit->plan->asleep);
		EnableItem(menu, miPlayDoneMoving);
		CheckItem(menu, miPlayDoneMoving, unit->plan->reserve);
	}
	if (can_create(unit) || can_complete(unit) /* or research? */) {
		EnableItem(menu, miPlayBuild);
	}
	if (can_repair(unit)) {
		EnableItem(menu, miPlayRepair);
	}
	if (can_attack(unit)) {
		EnableItem(menu, miPlayAttack);
	}
#if 0
	if (can_attack(unit)) {
		EnableItem(menu, miPlayOverrun);
	}
	if (can_attack(unit)) {
		EnableItem(menu, miPlayFire);
	}
#endif
	if (can_detonate(unit)) {
		EnableItem(menu, miPlayDetonate);
	}
	if (dside->designer) {
		EnableItem(menu, miPlayGive);
	}
	if (can_disband_at_all(dside, unit)) {
		EnableItem(menu, miPlayDisband);
	}
	if (1 /* !unitnameslocked */) {
		EnableItem(menu, miPlayRename);
	}
}

/* Keyboard command handling. */

do_keyboard_command(key)
char key;
{
	DGprintf("Typed '%c' (0x%x)\n", key, key);
	switch (key) {
		case 'a':
		case 'f':
			do_attack_command();
			break;
		case 's':
			do_sleep_command(1);
			break;
		case 'w':
			do_sleep_command(0);
			break;
		case '!':
			do_detonate_command();
			break;
		case '~':
			reset_coverage();
			break;
/*
		case '':
			do__mi();
			break;
*/
		default:
			break;
	}
}
