/* Copyright (c) 1992-1993 Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Random (single-instance) window handling for the Mac interface. */

#include "conq.h"
#include "mac.h"

char *unit_namer();
char *propose_unit_name();

/* Globals for the instructions window. */

WindowPtr instructionswin = nil;

TEHandle instructionstext = nil;

/* Globals for the game window. */

WindowPtr gamewin = nil;

int gamewinw = 200;
int gametoph = 20;
int gamesideh = 24;

Handle aisicnhandle = nil;
Handle facesicnhandle[3];

/* Globals for the construction window. */

WindowPtr buildwin = nil;

ListHandle buildunitlist = nil;
ListHandle buildtypelist = nil;

int constructmargin = 5;
int constructtop = 32;

int numposstypes, possibletypes[MAXUTYPES];
int numpossunits;
Unit *possibleunits[40];

int currunlength = 99;

ControlHandle constructbutton;
ControlHandle researchbutton;
ControlHandle runlengthpopup;

TEHandle runlengthtext = nil;

Rect unitlistrect, typelistrect;

/* Globals for the history window. */

WindowPtr historywin = nil;

/* Globals for the help window. */

DialogPtr helpwin = nil;

TEHandle helptext = nil;

struct a_helpnode *curhelpnode = NULL;
char *helpstring = NULL;
int helpscreentop = 0;

/* The help node that tells about how to use the help system. */

HelpNode *helphelpnode = NULL;

#define NODESTACKSIZE 50

HelpNode *nodestack[NODESTACKSIZE];

int nodestackpos;

/* The instructions window. */

pascal void
draw_instructions_text(win, ditem)
WindowPtr win;
short ditem;
{
	GrafPtr oldport;
	short itemtype;  Handle itemhandle;  Rect itemrect;

	GetDItem(instructionswin, diInstructionsText, &itemtype, &itemhandle, &itemrect);
	GetPort(&oldport);
	SetPort(instructionswin);
	TextSize(14);
	TEUpdate(&itemrect, instructionstext);
	TextSize(18);
	SetPort(oldport);
}

/* Bring up the dialog with instructions on how to play. */

instructions_dialog()
{
	if (instructionswin == nil) {
		create_instructions_dialog();
	}
	ShowWindow(instructionswin);
	SelectWindow(instructionswin);
}

create_instructions_dialog()
{
	Obj *instructions, *rest;
	char *str;
	Str255 tmpstr;
	Rect destrect, viewrect;
	short itemtype;  Handle itemhandle;  Rect itemrect;
	extern Module *selectedgame;

	instructionswin = GetNewDialog(dInstructions, nil, (DialogPtr) -1L);
	GetDItem(instructionswin, diInstructionsTitle, &itemtype, &itemhandle, &itemrect);
	c2p(((selectedgame && selectedgame->title) ? selectedgame->title : ""), tmpstr);
	SetIText(itemhandle, tmpstr);
	GetDItem(instructionswin, diInstructionsText, &itemtype, &itemhandle, &itemrect);
	SetDItem(instructionswin, diInstructionsText, itemtype, (Handle) draw_instructions_text, &itemrect);
	destrect = itemrect;
	viewrect = itemrect;
	SetPort(instructionswin);
	/* All text will be in Times. */
	/* (should put sizes in a resource so can be edited) */
	TextFont(times);
	TextSize(14);
	instructionstext = TENew(&destrect, &viewrect);
	TESetSelect(0, 100000, instructionstext);
	TEDelete(instructionstext);
	if (selectedgame && (instructions = selectedgame->instructions) != lispnil) {
		if (stringp(instructions)) {
			str = c_string(instructions);
			TEInsert(str, strlen(str), instructionstext);
		} else {
		    for (rest = instructions; rest != lispnil; rest = cdr(rest)) {
		    	if (stringp(car(rest))) {
		    		str = c_string(car(rest));
		    		if (strlen(str) == 0) {
		    			str = "\r\r";
		    		}
		    		TEInsert(str, strlen(str), instructionstext);
		    		TESetSelect(100000, 100000, instructionstext);
		    		TEInsert(" ", 1, instructionstext);
		    		TESetSelect(100000, 100000, instructionstext);
		    	}
	    	}
	    }
	} else {
		str = "(no instructions available)";
		TESetText(str, strlen(str), instructionstext);
	}
	TextSize(18);
	add_window_menu_item("Instructions", instructionswin);
}

#if 0

SideCloseup *
side_closeup_from_window(window)
WindowPtr window;
{
	SideCloseup *sidecloseup;
	
	for_all_side_closeups(sidecloseup) {
		if (sidecloseup->dialog == window) return sidecloseup;
	}
	return NULL;
}

draw_side_closeup(sidecloseup)
SideCloseup *sidecloseup;
{
	if (!active_display(dside) || sidecloseup == NULL) return;
	DrawDialog(sidecloseup->dialog);
}

hit_in_side_closeup(dialog, itemhit, evt)
DialogPtr dialog;
short itemhit;
EventRecord *evt;
{
	int closewhendone = FALSE;
	SideCloseup *sidecloseup = side_closeup_from_window(dialog);
	short itemtype;  Handle itemhandle;  Rect itemrect;
	
	if (sidecloseup == NULL) return FALSE;

	switch (itemhit) {
		case diSideCloseupMachineStrategy:
			GetDItem(dialog, diSideCloseupMachineStrategy, &itemtype, &itemhandle, &itemrect);
			SetCtlValue(itemhandle, !GetCtlValue(itemhandle));
			break;
		case diSideCloseupAutofinish:
			GetDItem(dialog, diSideCloseupAutofinish, &itemtype, &itemhandle, &itemrect);
			SetCtlValue(itemhandle, !GetCtlValue(itemhandle));
			break;
		case diSideCloseupOK:
			closewhendone = TRUE;
		case diSideCloseupApply:
			/* apply all the bits to the side. */
			GetDItem(dialog, diSideCloseupMachineStrategy, &itemtype, &itemhandle, &itemrect);
			dside->usemachinestrategy = GetCtlValue(itemhandle);
			GetDItem(dialog, diSideCloseupAutofinish, &itemtype, &itemhandle, &itemrect);
			dside->autofinish = GetCtlValue(itemhandle);
			break;
		case diSideCloseupCancel:
			closewhendone = TRUE;
			break;
	}
	if (closewhendone) {
		HideWindow(sidecloseup->dialog);
	}
	return TRUE;
}
#endif

/* The game progress window. */

/* Create the game progress window. */

create_game_window()
{
	/* Create the window, color if possible, since emblems may be in color. */
	if (hasColorQD) {	
		gamewin = GetNewCWindow(wGame, NULL, (WindowPtr) -1L);
	} else {
		gamewin = GetNewWindow(wGame, NULL, (WindowPtr) -1L);
	}
	if (keeping_score()) {
		/* need to recompute win width also */
		gamesideh += 20;
	}
	/* (should allocate extra display width if game is realtime) */
	/* This is not growable, so we have to ensure it's big enough to start with. */
	/* (should resize if more sides join) */
	SizeWindow(gamewin, gamewinw, numsides * gamesideh + gametoph, 1);
	/* Get handles to useful sicns. */
	aisicnhandle = GetNamedResource('SICN', "\pmplayer");
	facesicnhandle[0] = GetNamedResource('SICN', "\phostile");
	facesicnhandle[1] = GetNamedResource('SICN', "\pneutral");
	facesicnhandle[2] = GetNamedResource('SICN', "\pfriendly");
}

draw_game()
{
	Side *side2;
	GrafPtr oldport;

	if (gamewin == nil) return;
	GetPort(&oldport);
	SetPort(gamewin);
	EraseRect(&gamewin->portRect);
	draw_game_date();
	/* Draw a solid separating line between date info and side list. */
	MoveTo(0, gametoph);
	Line(gamewin->portRect.right, 0);
	for_all_sides(side2) {
		draw_game_side(side2);
	}
	SetPort(oldport);
}

/* Display the current time and date and any realtime countdowns. */

draw_game_date()
{
	Str255 tmpstr;
	Rect tmprect;

	SetRect(&tmprect, 0, 0, gamewin->portRect.right, gametoph - 1);
	EraseRect(&tmprect);
	c2p(absolute_date_string(curturn), tmpstr);
	MoveTo(10, 12);
	TextFace(bold);
	DrawString(tmpstr);
	TextFace(0);
	/* (should draw season name here) */
}

/* Draw info about a given side. */

draw_game_side(side2)
Side *side2;
{
	int sx = 20, sy = (side_number(side2) - 1) * gamesideh + gametoph;
	Str255 tmpstr;

	draw_side_emblem(gamewin, 2, sy + 4, 16, 16, side2);
	sprintf(spbuf, "%s", short_side_title(side2));
	c2p(spbuf, tmpstr);
	MoveTo(sx, sy + 12);
	/* Put the name of our side in boldface. */
	TextFace((side2 == dside ? bold : 0));
	DrawString(tmpstr);
	TextFace(0);
	if (side_has_ai(side2)) {
		/* Show that the side is run by an AI. */
		plot_sicn(gamewin, 182, sy + 2, aisicnhandle);
		/* what if side trusts but has no AI? */
		if (side2 != dside) {
			plot_sicn(gamewin, 164, sy + 2,
				facesicnhandle[feeling_towards(side2, dside)]);
		}
	}
	draw_side_status(side2);
	/* Draw a separating line. */
	PenPat(QD(gray));
	MoveTo(0, sy + gamesideh);
	Line(gamewin->portRect.right, 0);
	PenNormal();
}

feeling_towards(side, side2)
Side *side, *side2;
{
	if (side_has_ai(side) && should_try_to_win(side)) {
		return 0;
	} else {
		return 1;
	}
}

/* Draw the current details about a side. */

draw_side_status(side2)
Side *side2;
{
	int sx, sy = (side_number(side2) - 1) * gamesideh + gametoph, i, h, v;
	int totacp, resvacp, acpleft, percentleft, percentresv;
	Rect tmprect;

	if (!side2->ingame) {
		SetRect(&tmprect, 0, sy, gamewin->portRect.right, sy + gamesideh);
		gray_out_rect(&tmprect);
		if (side_won(side2)) {
			/* (should) Indicate that this side has won. */
			/* draw like a trophy or flourishes or some such?) */
		} else if (side_lost(side2)) {
			/* X out the loser. */
			PenPat(QD(ltGray));
			PenMode(patOr);
			PaintRect(&tmprect);
			PenNormal();
		}
	} else {
		/* Show the current acp totals/progress of the side. */
		/* This is not quite the security hole it might seem,
		   you don't get much value out of seeing how far along each side is. */
		SetRect(&tmprect, 20, sy + 12 + 4, 20 + 100, sy + 12 + 4 + 7);
		EraseRect(&tmprect);
		totacp = side_initacp(side2);
		if (totacp > 0) {
			FrameRect(&tmprect);
			acpleft = side_acp(side2);
			resvacp = side_acp_reserved(side2);
			if (totacp > 0) {
				percentleft = (100 * acpleft) / totacp;
				percentleft = max(0, min(99, percentleft));
				percentresv = (100 * resvacp) / totacp;
				percentresv = max(0, min(99, percentresv));
			} else {
				percentleft = percentresv = 0;
			}
			if (percentleft > 0) {
				InsetRect(&tmprect, 1, 1);
				tmprect.right = tmprect.left + percentleft;
				FillRect(&tmprect, QD(black));
			}
			if (percentresv > 0) {
				PenPat(QD(gray));
				MoveTo(20 + percentresv, sy + 12 + 4);
				Line(0, 6);
				PenNormal();
			}
		} else {
			/* Draw a gray frame to indicate that this side has no units
			   that can actually do anything. */
			PenPat(QD(gray));
			FrameRect(&tmprect);
			PenNormal();
		}
		if (side2->finishedturn) {
			/* should do something more obvious */
			SetRect(&tmprect, 20, sy + 12 + 4, 20 + 100, 12 + 4 + 7);
			InsetRect(&tmprect, 1, 1);
			gray_out_rect(&tmprect);
		}
	}
	if (keeping_score()) {
		/* (should draw known values of each score here) */
	}
}

/* The construction planning window. */

create_construction_window()
{
	int done = FALSE, u, num, wid, hgt;
	short ditem, i;
	Unit *unit;
	Point pnt, cellsize;
	Cell tmpcell;
	Rect listrect, destrect, viewrect, tmprect;
	short itemtype;  Handle itemhandle;  Rect itemrect;

	if (hasColorQD) {
		buildwin = GetNewCWindow(wConstruction, NULL, (WindowPtr) -1L);
	} else {
		buildwin = GetNewWindow(wConstruction, NULL, (WindowPtr) -1L);
	}
	constructbutton = GetNewControl(cConstructButton, buildwin);
	researchbutton = GetNewControl(cResearchButton, buildwin);
	runlengthpopup = GetNewControl(210, buildwin);
	SetPort(buildwin);
	TextFont(monaco);
	TextSize(9);
	/* Set up the user item that lists all the units. */
	calc_construction_rects();
	SetRect(&listrect, 0, 0, 1, 0);
	SetPt(&cellsize, 0, 0);
	/* Create the list of units itself. */
	tmprect = unitlistrect;
	tmprect.right -= sbarwid;
	buildunitlist = LNew(&tmprect, &listrect, cellsize, 0, buildwin,
						 FALSE, FALSE, FALSE, TRUE);
	/* Now set up the list of types. */
	SetRect(&listrect, 0, 0, 1, 0);
	SetPt(&cellsize, 0, 0);
	tmprect = typelistrect;
	tmprect.right -= sbarwid;
	buildtypelist = LNew(&tmprect, &listrect, cellsize, 0, buildwin,
						 FALSE, FALSE, FALSE, TRUE);
	init_construction_lists();
	add_window_menu_item("Construction", buildwin);
	/* (should keep the window out of the way of the frontmost map?) */
	ShowWindow(buildwin);
}

init_construction_lists()
{
	int u, tm, num;
	Unit *unit;
	Point cellsize;
	Cell tmpcell, tmpcell2;
	Rect listrect, destrect, viewrect;

	/* Update the list of units. */
	LDoDraw(0, buildunitlist);
	LDelRow(0, 0, buildunitlist);
	SetPt(&tmpcell, 0, 0);
/*	LCellSize(tmpcell, buildunitlist); */
	numpossunits = 0;
	for_all_side_units(dside, unit) {
		if (is_acting(unit) && can_build(unit) && numpossunits < 40) {
			/* (should be a separate routine?) */
			LAddRow(1, tmpcell.v, buildunitlist);
			construction_description(unit);
			LSetCell(spbuf, strlen(spbuf), tmpcell, buildunitlist);
			++tmpcell.v;
			possibleunits[numpossunits++] = unit;
		}
	}
	LDoDraw(1, buildunitlist);
	/* Update the list of types. */
	LDoDraw(0, buildtypelist);
	LDelRow(0, 0, buildtypelist);
	SetPt(&tmpcell, 0, 0);
/*	LCellSize(tmpcell, buildtypelist); */
	numposstypes = 0;
	for_all_unit_types(u) {
		if (1 /* could be built by some unit in the list */) {
			LAddRow(1, tmpcell.v, buildtypelist);
			sprintf(spbuf, "     %s", u_type_name(u));
			if ((num = num_units_in_play(dside, u)) > 0) {
				sprintf(spbuf+strlen(spbuf), "  %d", num);
			}
			if ((num = num_units_incomplete(dside, u)) > 0) {
				sprintf(spbuf+strlen(spbuf), " (%d)", num);
			}
			LSetCell(spbuf, strlen(spbuf), tmpcell, buildtypelist);
			++tmpcell.v;
			possibletypes[numposstypes++] = u;
		}
	}
	LDoDraw(1, buildtypelist);
}

/* (should go to nlang.c?) */

construction_description(unit)
Unit *unit;
{
	Task *task;
	Unit *unit2;

	sprintf(spbuf, "     %s", u_type_name(unit->type));
	if (unit->name) {
		sprintf(spbuf+strlen(spbuf), " %s", unit->name);
	}
	if (unit->plan && unit->plan->tasks && unit->plan->tasks->type == BUILD_TASK) {
		task = unit->plan->tasks;		
		sprintf(spbuf+strlen(spbuf), "  %s (", u_type_name(task->args[0]));
		unit2 = find_unit(task->args[1]);
		if (in_play(unit2) && unit2->type == task->args[0]) {
			sprintf(spbuf+strlen(spbuf), "%d/%d done, ", unit2->cp, u_cp(unit2->type));
		}
		sprintf(spbuf+strlen(spbuf), "%d of %d)",
				task->args[2], task->args[3]);
	}
}

est_completion_time(unit, u2)
Unit *unit;
int u2;
{
	if (uu_acp_to_create(unit->type, u2) < 1) return (-1);
	return normal_completion_time(unit->type, u2);
}

draw_construction()
{
	Rect tmprect;

	calc_construction_rects();
	LUpdate(buildwin->visRgn, buildunitlist);
	tmprect = unitlistrect;
	InsetRect(&tmprect, -1, -1);
	FrameRect(&tmprect);
	LUpdate(buildwin->visRgn, buildtypelist);
	tmprect = typelistrect;
	InsetRect(&tmprect, -1, -1);
	FrameRect(&tmprect);
}

calc_construction_rects()
{
	int wid, hgt, divide;
	Rect tmprect;

	tmprect = buildwin->portRect;
	wid = tmprect.right - tmprect.left - sbarwid;
	hgt = tmprect.bottom - tmprect.top - sbarwid;
	divide = (wid * 2) / 3;
	SetRect(&unitlistrect, 0, constructtop, divide, hgt);
	InsetRect(&unitlistrect, constructmargin, constructmargin);
	SetRect(&typelistrect, divide, constructtop, wid, hgt);
	InsetRect(&typelistrect, constructmargin, constructmargin);
}

Unit *
get_selected_construction_unit()
{
	Point tmpcell;
	Unit *unit;

	SetPt(&tmpcell, 0, 0);
	if (LGetSelect(TRUE, &tmpcell, buildunitlist)) {				
		if (tmpcell.v < numpossunits) {
			unit = possibleunits[tmpcell.v];
			if (is_acting(unit)) return unit;
		}
	}
	return NULL;
}

get_selected_construction_type()
{
	int u;
	Point tmpcell;

	SetPt(&tmpcell, 0, 0);
	if (LGetSelect(TRUE, &tmpcell, buildtypelist)) {				
		if (tmpcell.v < numposstypes) {
			return possibletypes[tmpcell.v];
		}
	}
	return NONUTYPE;
}

do_mouse_down_construction(mouse, mods)
Point mouse;
int mods;
{
	ControlHandle control;
	short part, value;
	int u;
	Point pt;
	Unit *unit;

	part = FindControl(mouse, buildwin, &control);
	if (control == constructbutton) {
		if ((unit = get_selected_construction_unit()) != NULL) {
			if ((u = get_selected_construction_type()) != NONUTYPE) {
				push_build_task(unit, u, currunlength);
				update_construction_unit_list(unit);
				return;
			}
		}
		SysBeep(20);  /* should never happen? */
	} else if (control == researchbutton) {
		/* should set up a research task? */
	} else if (control == runlengthpopup) {
		TrackControl(control, mouse, (void *) -1);
		currunlength = GetCtlValue(control);
	} else if (PtInRect(mouse, &unitlistrect)) {
		LClick(mouse, mods, buildunitlist);
		/* Update the type list to show what could be built and in how long. */
		update_type_list_for_unit(get_selected_construction_unit());
	} else if (PtInRect(mouse, &typelistrect)) {
		LClick(mouse, mods, buildtypelist);
		/* Update the unit list to show what could build the type */
		update_unit_list_for_type(get_selected_construction_type());
	} else {
		/* Click was not in any useful part of the window. */ 
	}
}

/* Highlight exactly one specific unit in the build dialog, and unhighlight
   any others. */

select_unit_in_construction_window(unit)
Unit *unit;
{
	int i;
	Point tmpcell;

	for (i = 0; i < numpossunits; ++i) {
		SetPt(&tmpcell, 0, i);
		LSetSelect((unit == possibleunits[i]), tmpcell, buildunitlist);
	}
	update_type_list_for_unit(get_selected_construction_unit());
}

/* Given a unit (which may be any unit), update the list of constructing units. */

update_construction_unit_list(unit)
Unit *unit;
{
	int i;
	Point tmpcell;

	for (i = 0; i < numpossunits; ++i) {
		if (unit == possibleunits[i]) {
			SetPt(&tmpcell, 0, i);
			if (is_acting(unit) && unit->side == dside) {
				construction_description(unit);
				LSetCell(spbuf, strlen(spbuf), tmpcell, buildunitlist);
			} else {
/*				LDelRow(1, tmpcell.v, buildunitlist); */
				LSetCell("", 0, tmpcell, buildunitlist);
			}
			return;
		}
	}
	/* Unit not found, maybe add to the list. */
	if (unit->side == dside) {
		if (is_acting(unit) && can_build(unit) && numpossunits < 40) {
			SetPt(&tmpcell, 0, numpossunits);
			LAddRow(1, numpossunits, buildunitlist);
			construction_description(unit);
			LSetCell(spbuf, strlen(spbuf), tmpcell, buildunitlist);
			possibleunits[numpossunits++] = unit;
		}
	}
}

update_unit_list_for_type(u)
int u;
{
}

update_type_list_for_unit(unit)
Unit *unit;
{
#if 0
	int tm;
	Point tmpcell2;

	SetPt(&tmpcell2, 0, 0);
	if (LGetSelect(TRUE, &tmpcell2, buildtypelist)
	    && tmpcell2.v < numposstypes
	    && ((tm = est_completion_time(unit, possibletypes[tmpcell2.v])) >= 0)) {
		sprintf(buf, "[%2d]", tm);
	} else {
		sprintf(buf, "    ");
	}
#endif
}

/* Resize the construction window to the given size. */

grow_construction(winsize)
long winsize;
{
	int wid, hgt, leftwid;
	Rect tmprect;

	EraseRect(&buildwin->portRect);
	SizeWindow(buildwin, LoWord(winsize), HiWord(winsize), 1);
	/* Recalculate size and position of the unit and type lists. */
	calc_construction_rects();
	LSize(unitlistrect.right - unitlistrect.left - sbarwid,
		  unitlistrect.bottom - unitlistrect.top,
		  buildunitlist);
	/* Move the type list (is this the approved way to do it?) */
	(*buildtypelist)->rView.left = typelistrect.left;
	LSize(typelistrect.right - typelistrect.left - sbarwid,
		  typelistrect.bottom - typelistrect.top,
		  buildtypelist);
	/* This will force a full redraw at the next update. */
	InvalRect(&buildwin->portRect);
}					

/* The side closeup window. */

create_side_closeup(side2)
Side *side2;
{
	SideCloseup *sidecloseup = (SideCloseup *) xmalloc(sizeof(SideCloseup));
	Rect vscrollrect, hscrollrect;

	DGprintf("Creating a side closeup\n");
	sidecloseup->side = side2;
	sidecloseup->dialog = GetNewDialog(dSideCloseup, nil, (DialogPtr) -1L);
	sidecloseup->next = sidecloseuplist;
	sidecloseuplist = sidecloseup;
	ShowWindow(sidecloseup->dialog);
	add_window_menu_item("Side Closeup", sidecloseup->dialog);
}


SideCloseup *
side_closeup_from_window(window)
WindowPtr window;
{
	SideCloseup *sidecloseup;
	
	for_all_side_closeups(sidecloseup) {
		if (sidecloseup->dialog == window) return sidecloseup;
	}
	return NULL;
}

draw_side_closeup(sidecloseup)
SideCloseup *sidecloseup;
{
	if (!active_display(dside) || sidecloseup == NULL) return;
	DrawDialog(sidecloseup->dialog);
}

hit_in_side_closeup(dialog, itemhit, evt)
DialogPtr dialog;
short itemhit;
EventRecord *evt;
{
	int closewhendone = FALSE;
	SideCloseup *sidecloseup = side_closeup_from_window(dialog);
	short itemtype;  Handle itemhandle;  Rect itemrect;
	
	if (sidecloseup == NULL) return FALSE;

	switch (itemhit) {
		case diSideCloseupMachineStrategy:
			GetDItem(dialog, diSideCloseupMachineStrategy, &itemtype, &itemhandle, &itemrect);
			SetCtlValue((ControlHandle) itemhandle, !GetCtlValue((ControlHandle) itemhandle));
			break;
		case diSideCloseupAutofinish:
			GetDItem(dialog, diSideCloseupAutofinish, &itemtype, &itemhandle, &itemrect);
			SetCtlValue((ControlHandle) itemhandle, !GetCtlValue((ControlHandle) itemhandle));
			break;
		case diSideCloseupOK:
			closewhendone = TRUE;
		case diSideCloseupApply:
			/* apply all the bits to the side. */
			GetDItem(dialog, diSideCloseupMachineStrategy, &itemtype, &itemhandle, &itemrect);
			dside->usemachinestrategy = GetCtlValue((ControlHandle) itemhandle);
			GetDItem(dialog, diSideCloseupAutofinish, &itemtype, &itemhandle, &itemrect);
			dside->autofinish = GetCtlValue((ControlHandle) itemhandle);
			break;
		case diSideCloseupCancel:
			closewhendone = TRUE;
			break;
	}
	if (closewhendone) {
		HideWindow(sidecloseup->dialog);
	}
	return TRUE;
}

char *
get_string_from_item(itemhandle)
Handle itemhandle;
{
	char tmpbuf[BUFSIZE];
	Str255 tmpstr;
	
	GetIText(itemhandle, tmpstr);
	/* This is basically p2c. */
 	strncpy(tmpbuf, tmpstr+1, tmpstr[0]);
	/* Make sure the string is terminated properly. */
	tmpbuf[tmpstr[0]] = '\0';
	return copy_string(tmpbuf);
}

side_rename_dialog(side)
Side *side;
{
	short done = FALSE, changed = TRUE, ditem;
	Str255 tmpstr;
	DialogPtr win;
	short itemtype;  Handle itemhandle;  Rect itemrect;

	win = GetNewDialog(dSideRename, NULL, (DialogPtr) -1L);
	while (!done) {
		if (changed) {
			/* Seed the items with the current side names. */
			GetDItem(win, diSideRenameName, &itemtype, &itemhandle, &itemrect);
			c2p((side->name ? side->name : ""), tmpstr);
			SetIText(itemhandle, tmpstr);
			GetDItem(win, diSideRenameFullName, &itemtype, &itemhandle, &itemrect);
			c2p((side->longname ? side->longname : ""), tmpstr);
			SetIText(itemhandle, tmpstr);
			GetDItem(win, diSideRenameAcronym, &itemtype, &itemhandle, &itemrect);
			c2p((side->shortname ? side->shortname : ""), tmpstr);
			SetIText(itemhandle, tmpstr);
			GetDItem(win, diSideRenameNoun, &itemtype, &itemhandle, &itemrect);
			c2p((side->noun ? side->noun : ""), tmpstr);
			SetIText(itemhandle, tmpstr);
			GetDItem(win, diSideRenamePluralNoun, &itemtype, &itemhandle, &itemrect);
			c2p((side->pluralnoun ? side->pluralnoun : ""), tmpstr);
			SetIText(itemhandle, tmpstr);
			GetDItem(win, diSideRenameAdjective, &itemtype, &itemhandle, &itemrect);
			c2p((side->adjective ? side->adjective : ""), tmpstr);
			SetIText(itemhandle, tmpstr);
			GetDItem(win, diSideRenameEmblemName, &itemtype, &itemhandle, &itemrect);
			c2p((side->emblemname ? side->emblemname : ""), tmpstr);
			SetIText(itemhandle, tmpstr);
			GetDItem(win, diSideRenameColorScheme, &itemtype, &itemhandle, &itemrect);
			c2p((side->colorscheme ? side->colorscheme : ""), tmpstr);
			SetIText(itemhandle, tmpstr);
			ShowWindow(win);
			changed = FALSE;
		}
		ModalDialog(NULL, &ditem);
		switch (ditem) {
			case diSideRenameOK:
				/* Actually change the side's slots. */
				GetDItem(win, diSideRenameName, &itemtype, &itemhandle, &itemrect);
				side->name = get_string_from_item(itemhandle);
				GetDItem(win, diSideRenameFullName, &itemtype, &itemhandle, &itemrect);
				side->longname = get_string_from_item(itemhandle);
				GetDItem(win, diSideRenameAcronym, &itemtype, &itemhandle, &itemrect);
				side->shortname = get_string_from_item(itemhandle);
				GetDItem(win, diSideRenameNoun, &itemtype, &itemhandle, &itemrect);
				side->noun = get_string_from_item(itemhandle);
				GetDItem(win, diSideRenamePluralNoun, &itemtype, &itemhandle, &itemrect);
				side->pluralnoun = get_string_from_item(itemhandle);
				GetDItem(win, diSideRenameAdjective, &itemtype, &itemhandle, &itemrect);
				side->adjective = get_string_from_item(itemhandle);
				GetDItem(win, diSideRenameEmblemName, &itemtype, &itemhandle, &itemrect);
				side->emblemname = get_string_from_item(itemhandle);
				GetDItem(win, diSideRenameColorScheme, &itemtype, &itemhandle, &itemrect);
				side->colorscheme = get_string_from_item(itemhandle);
				/* Force redisplay of everything that might use any side names. */
				force_overall_update();
				/* Fall into next case. */
			case diSideRenameCancel:
				done = TRUE;
				break;
			case diSideRenameRandom:
				side->name = NULL;
				side->noun = NULL;
				/* always need to clear this cache before renaming... */
				side->pluralnoun = NULL;
				side->adjective = NULL;
				make_up_side_name(side);
				init_emblem_images(side);
				changed = TRUE;
				break;
		}
	}
	DisposDialog(win);
}

/* Generic renaming, applies to both units and features. */

unit_rename_dialog(unit)
Unit *unit;
{
	short done = FALSE, ditem;
	char *newname;
	char *namer = unit_namer(unit);
	Str255 tmpstr;
	DialogPtr win;
	short itemtype;  Handle itemhandle;  Rect itemrect;

	if (unit == NULL) return;
	win = GetNewDialog(dRename, NULL, (DialogPtr) -1L);
	/* Seed the text item with the original name. */
	newname = unit->name;
	if (newname == NULL) newname = "";
	GetDItem(win, diRenameName, &itemtype, &itemhandle, &itemrect);
	c2p(newname, tmpstr);
	SetIText(itemhandle, tmpstr);
	/* Gray out the random renaming button if no namers available. */
	GetDItem(win, diRenameRandom, &itemtype, &itemhandle, &itemrect);
	HiliteControl((ControlHandle) itemhandle, ((namer != NULL) ? 0 : 255));
	ShowWindow(win);
	while (!done) {
		ModalDialog(NULL, &ditem);
		switch (ditem) {
			case diRenameOK:
				GetDItem(win, diRenameName, &itemtype, &itemhandle, &itemrect);
				newname = get_string_from_item(itemhandle);
				/* Turn 0-length names into NULL, which usually means "unnamed". */
				if (newname != NULL && strlen(newname) == 0) newname = NULL;
				unit->name = newname;
				update_unit_display(dside, unit, TRUE);
				/* (which other sides should also see the new name?) */
				/* Fall into next case. */
			case diRenameCancel:
				done = TRUE;
				break;
			case diRenameRandom:
				newname = propose_unit_name(unit);
				GetDItem(win, diRenameName, &itemtype, &itemhandle, &itemrect);
				c2p(newname, tmpstr);
				SetIText(itemhandle, tmpstr);
				break;
		}
	}
	DisposDialog(win);
}

UnitCloseup *
find_unit_closeup(unit)
Unit *unit;
{
	UnitCloseup *unitcloseup;

	for_all_unit_closeups(unitcloseup) {
		if (unitcloseup->unit == unit
			&& unitcloseup->window
			&& ((WindowPeek) unitcloseup->window)->visible) return unitcloseup;
	}
	return NULL;
}

#if 0

pascal void
draw_unit_closeup_image(win, ditem)
WindowPtr win;
short ditem;
{
	short itemtype;  Handle itemhandle;  Rect itemrect;
	UnitCloseup *unitcloseup = (UnitCloseup *) unit_closeup_from_window(win);

	GetDItem(win, diUnitCloseupImage, &itemtype, &itemhandle, &itemrect);
	EraseRect(&itemrect);
	if (unitcloseup && unitcloseup->unit) {
		draw_unit_image(win, itemrect.left, itemrect.top,
						itemrect.right - itemrect.left, itemrect.bottom - itemrect.top,
						unitcloseup->unit->type, side_number(unitcloseup->unit->side));
	} else {
		FrameRect(&itemrect);
	}
}

pascal void
draw_unit_closeup_supply(win, ditem)
WindowPtr win;
short ditem;
{
	int m;
	Unit *unit;
	short itemtype;  Handle itemhandle;  Rect itemrect;
	UnitCloseup *unitcloseup = (UnitCloseup *) unit_closeup_from_window(win);
	Str255 tmpstr;

	GetDItem(win, diUnitCloseupSupply, &itemtype, &itemhandle, &itemrect);
	EraseRect(&itemrect);
	if (unitcloseup && (unit = unitcloseup->unit) != NULL && has_any_storage(unit)) {
		/* Show the stored materials. */
		for_all_material_types(m) {
			MoveTo(itemrect.left, itemrect.top + 10 + 15 * m);
			sprintf(spbuf, "%s %d/%d",
					m_type_name(m), unit->supply[m], um_storage(unit->type, m));
			c2p(spbuf, tmpstr);
			DrawString(tmpstr);
		}
	}
}

has_any_storage(unit)
Unit *unit;
{
	int m;

	for_all_material_types(m) {
		if (um_storage(unit->type, m) > 0) return TRUE;
	}
	return FALSE;
}

pascal void
draw_unit_closeup_tooling(win, ditem)
WindowPtr win;
short ditem;
{
	Unit *unit;
	UnitCloseup *unitcloseup = (UnitCloseup *) unit_closeup_from_window(win);
	short itemtype;  Handle itemhandle;  Rect itemrect;

	GetDItem(win, diUnitCloseupTooling, &itemtype, &itemhandle, &itemrect);
	EraseRect(&itemrect);
	if (unitcloseup && ((unit = unitcloseup->unit) != NULL)) {
		/* (eventually display tooling) */
	} else {
	}
}

pascal void
draw_unit_closeup_occupants(win, ditem)
WindowPtr win;
short ditem;
{
	int i;
	Str255 tmpstr;
	Unit *unit, *occ;
	UnitCloseup *unitcloseup = (UnitCloseup *) unit_closeup_from_window(win);
	short itemtype;  Handle itemhandle;  Rect itemrect;

	GetDItem(win, diUnitCloseupOccs, &itemtype, &itemhandle, &itemrect);
	EraseRect(&itemrect);
	if (unitcloseup && ((unit = unitcloseup->unit) != NULL) && unit->occupant) {
		FrameRect(&itemrect);
		i = 0;
		for_all_occupants(unit, occ) {
			MoveTo(itemrect.left, itemrect.top + 10 + 15 * i);
			sprintf(spbuf, "%s", unit_desig(occ));
			c2p(spbuf, tmpstr);
			DrawString(tmpstr);
			++i;
		}
	} else {
	}
}

#endif

create_unit_closeup(unit)
Unit *unit;
{
	int u;
	Str255 tmpstr;
	WindowPtr win;
	UnitCloseup *unitcloseup = (UnitCloseup *) xmalloc(sizeof(UnitCloseup));
	short itemtype;  Handle itemhandle;  Rect itemrect;

	if (!active_display(dside) || unit == NULL) return;
	DGprintf("Creating a closeup of %s\n", unit_desig(unit));
	u = unit->type;
	unitcloseup->unit = unit;
	if (hasColorQD) {
		win = GetNewCWindow(wUnitCloseup, nil, (WindowPtr) -1L);
	} else {
		win = GetNewWindow(wUnitCloseup, nil, (WindowPtr) -1L);
	}
	unitcloseup->window = win;
	unitcloseup->next = unitcloseuplist;
	unitcloseuplist = unitcloseup;
	SizeWindow(win, 200, (unit->act ? 200 :100), 1);
	sprintf(spbuf, "%s %d Closeup", u_type_name(u), unit->id);
	add_window_menu_item(spbuf, win);
	/* We're now ready to show this closeup to the world. */
	ShowWindow(win);
}

draw_unit_closeup(unitcloseup)
UnitCloseup *unitcloseup;
{
	int u, m;
	Str255 tmpstr;
	Rect tmprect;
	GrafPtr oldport;
	WindowPtr win = unitcloseup->window;
	Unit *unit = unitcloseup->unit;

	if (!active_display(dside)) return;
	if (unit == NULL || !alive(unit)) {
		/* should close the window permanently */
		unitcloseup->window = nil;
		HideWindow(win);
		return;
	}
	GetPort(&oldport);
	SetPort(win);
	EraseRect(&win->portRect);
	u = unit->type;
	/* Draw the unit's image. */
	/* (Draw current terrain underneath also?) */
	SetRect(&tmprect, 5, 5, 37, 37); 
	EraseRect(&tmprect);
	draw_unit_image(win, tmprect.left, tmprect.top,
					tmprect.right - tmprect.left, tmprect.bottom - tmprect.top,
					u, side_number(unit->side), !completed(unit));
	/* Draw the unit's name. */
	if (unit->name) {
		c2p(unit->name, tmpstr);
	} else if (unit->number > 0) {
		sprintf(tmpbuf, "%d%s", unit->number, ordinal_suffix(unit->number));
		c2p(tmpbuf, tmpstr);
	} else {
		c2p("", tmpstr);
	}
	MoveTo(45, 21);
	DrawString(tmpstr);
	c2p(side_and_type_name(dside, u, unit->side), tmpstr);
	MoveTo(45, 41);
	DrawString(tmpstr);
	if (unit->z != 0) {
		sprintf(tmpbuf, "; alt %d", unit->z);
	} else {
		strcpy(tmpbuf, "");
	}
	sprintf(spbuf, "at %d,%d%s", unit->x, unit->y, tmpbuf);
	if (unit->transport != NULL) {
		sprintf(tmpbuf, "In %s (%s)", unit_desig(unit->transport), spbuf);
		c2p(tmpbuf, tmpstr);
	} else {
		c2p(spbuf, tmpstr);
	}
	MoveTo(5, 61);
	DrawString(tmpstr);
	if (completed(unit)) {
		sprintf(tmpbuf, "HP %d/%d", unit->hp, u_hp(u));
	} else {
		sprintf(tmpbuf, "%d/%d done", unit->cp, u_cp(u));
	}
	c2p(tmpbuf, tmpstr);
	MoveTo(5, 81);
	DrawString(tmpstr);
	/* Display the acp. */
	if (completed(unit) && unit->act) {
		sprintf(tmpbuf, "ACP %d/%d", unit->act->acp, unit->act->initacp);
	} else {
		strcpy(tmpbuf, "");
	}
	c2p(tmpbuf, tmpstr);
	MoveTo(5, 101);
	DrawString(tmpstr);
	/* Display the unit's supplies. */
	sprintf(tmpbuf, "");
	for_all_material_types(m) {
		sprintf(tmpbuf+strlen(tmpbuf), "%s %d  ", m_type_name(m), unit->supply[m]);
	}
	c2p(tmpbuf, tmpstr);
	MoveTo(5, 121);
	DrawString(tmpstr);
	if (unit->plan) {
		int i = 161;
		Task *task;
		extern char *plantypenames[];
		Plan *plan = unit->plan;
		
		/* Should be able to say more about the plan */
		sprintf(tmpbuf, "plan %s", plantypenames[plan->type]);
    	if (plan->waitingfortasks) strcat(tmpbuf, " [waiting]");
    	if (plan->reserve) strcat(tmpbuf, " [reserve]");
    	if (plan->asleep) strcat(tmpbuf, " [asleep]");
		c2p(tmpbuf, tmpstr);
		MoveTo(5, 141);
		DrawString(tmpstr);
		if (plan->tasks) {
			for (task = plan->tasks; task != NULL; task = task->next) {
				sprintf(tmpbuf, "task %s", task_desig(task));
				c2p(tmpbuf, tmpstr);
				MoveTo(5, i);
				DrawString(tmpstr);
				i += 20;			
			}
		}
	}
	SetPort(oldport);
}


UnitCloseup *
unit_closeup_from_window(win)
WindowPtr win;
{
	UnitCloseup *unitcloseup;
	
	for_all_unit_closeups(unitcloseup) {
		if (unitcloseup->window == win) return unitcloseup;
	}
	return NULL;
}

#if 0

hit_in_unit_closeup(dialog, itemhit, evt)
DialogPtr dialog;
short itemhit;
EventRecord *evt;
{
	UnitCloseup *unitcloseup = unit_closeup_from_window(dialog);
	short itemtype;  Handle itemhandle;  Rect itemrect;
	
	if (unitcloseup == NULL) return FALSE;

#if 0
	switch (itemhit) {
		case diSideCloseupMachineStrategy:
			GetDItem(dialog, diSideCloseupMachineStrategy, &itemtype, &itemhandle, &itemrect);
			SetCtlValue(itemhandle, !GetCtlValue(itemhandle));
			break;
		case diSideCloseupAutofinish:
			GetDItem(dialog, diSideCloseupAutofinish, &itemtype, &itemhandle, &itemrect);
			SetCtlValue(itemhandle, !GetCtlValue(itemhandle));
			break;
		case diSideCloseupOK:
			closewhendone = TRUE;
		case diSideCloseupApply:
			/* apply all the bits to the side. */
			GetDItem(dialog, diSideCloseupMachineStrategy, &itemtype, &itemhandle, &itemrect);
			dside->usemachinestrategy = GetCtlValue(itemhandle);
			GetDItem(dialog, diSideCloseupAutofinish, &itemtype, &itemhandle, &itemrect);
			dside->autofinish = GetCtlValue(itemhandle);
			break;
		case diSideCloseupCancel:
			closewhendone = TRUE;
			break;
	}
#endif
	update_unit_display(dside, unitcloseup->unit, TRUE);
	return TRUE;
}

#endif

/* History window. */

create_history_window()
{
	historywin = GetNewWindow(wHistory, NULL, (WindowPtr) -1L);
}

int histi;

draw_history()
{
	GrafPtr oldport;
    HistEvent *hevt;

	if (!active_display(dside) || historywin == nil) return;

	GetPort(&oldport);
	SetPort(historywin);
	histi = 0;
	display_historical_event(history);
	for (hevt = history->next; hevt != history; hevt = hevt->next) {
		++histi;
    	display_historical_event(hevt);
	}
	SetPort(oldport);
}

display_historical_event(hevt)
HistEvent *hevt;
{
	char hevtbuf[500];
	Str255 tmpstr;

	MoveTo(2, 15 * histi + 10);
	sprintf(tmpbuf, "%d: ", hevt->startdate);
	c2p(tmpbuf, tmpstr);
	DrawString(tmpstr);
	MoveTo(40, 15 * histi + 10);
	historical_event_desc(hevt, hevtbuf);
	c2p(hevtbuf, tmpstr);
	DrawString(tmpstr);
}

/* Help dialog. */

/* This is the top-level access to help, can be called anywhere, anytime. */

help_dialog()
{
	if (helpwin == nil) {
		create_help_dialog();
	}
	ShowWindow(helpwin);
	SelectWindow(helpwin);
}

describe_menus(arg, key, buf)
int arg;
char *key, *buf;
{
	sprintf(buf, "[describe the menus]");
}

describe_mouse(arg, key, buf)
int arg;
char *key, *buf;
{
	sprintf(buf, "Click once on a unit to select it");
}

describe_shortcuts(arg, key, buf)
int arg;
char *key, *buf;
{
	sprintf(buf, "[describe any shortcuts]");
}

describe_help(arg, key, buf)
int arg;
char *key, *buf;
{
	sprintf(buf, "Click on the arrows");
}

pascal void
draw_help_image(win, ditem)
WindowPtr win;
short ditem;
{
	GrafPtr oldport;
	short itemtype;  Handle itemhandle;  Rect itemrect;

	GetDItem(helpwin, diHelpIcon, &itemtype, &itemhandle, &itemrect);
	GetPort(&oldport);
	SetPort(helpwin);
	if (is_unit_type(curhelpnode->arg)) {
		draw_unit_image(helpwin, itemrect.left, itemrect.top,
						itemrect.right - itemrect.left, itemrect.bottom - itemrect.top,
						curhelpnode->arg, -1, 0);
	}
	SetPort(oldport);
}

pascal void
draw_help_text(win, ditem)
WindowPtr win;
short ditem;
{
	GrafPtr oldport;
	short itemtype;  Handle itemhandle;  Rect itemrect;

	GetDItem(helpwin, diHelpText, &itemtype, &itemhandle, &itemrect);
	GetPort(&oldport);
	SetPort(helpwin);
	TextSize(14);
	TEUpdate(&itemrect, helptext);
	TextSize(18);
	SetPort(oldport);
}

create_help_dialog()
{
	Rect destrect, viewrect;
	short itemtype;  Handle itemhandle;  Rect itemrect;

	helpwin = GetNewDialog(dHelp, NULL, (DialogPtr) -1L);
	GetDItem(helpwin, diHelpIcon, &itemtype, &itemhandle, &itemrect);
	SetDItem(helpwin, diHelpIcon, itemtype, (Handle) draw_help_image, &itemrect);
	GetDItem(helpwin, diHelpText, &itemtype, &itemhandle, &itemrect);
	SetDItem(helpwin, diHelpText, itemtype, (Handle) draw_help_text, &itemrect);
	destrect = itemrect;
	viewrect = itemrect;
	SetPort(helpwin);
	/* All text will be in Times. */
	/* (should these be choosable?) */
	TextFont(times);
	TextSize(14);
	helptext = TENew(&destrect, &viewrect);
	TextSize(18);
	/* Add the Mac-specific help nodes. */
	helphelpnode = add_help_node("help", describe_help, 0, firsthelpnode);
	add_help_node("shortcuts", describe_shortcuts, 0, firsthelpnode);
	add_help_node("mouse input", describe_mouse, 0, firsthelpnode);
	add_help_node("menus", describe_menus, 0, firsthelpnode);
	curhelpnode = firsthelpnode;
	set_help_content(curhelpnode);
	nodestackpos = 0;
	add_window_menu_item("Help", helpwin);
}

set_help_content(curnode)
HelpNode *curnode;
{
	int i;
	char *str;
	Str255 tmpstr;
	short itemtype;  Handle itemhandle;  Rect itemrect;

	get_help_text(curnode);
	/* Set the displayed topic. */
	GetDItem(helpwin, diHelpTopic, &itemtype, &itemhandle, &itemrect);
	c2p(curnode->key, tmpstr);
	SetIText(itemhandle, tmpstr);
	/* Set the displayed text. */
	str = curnode->text;
	/* Hack up newlines so that TextEdit recognizes them. */
	for (i = 0; i < strlen(str); ++i) {
		if (str[i] == '\n') str[i] = '\r';
	}
	helpstring = str;
	/* Remove all the existing text. */
	TESetSelect(0, 100000, helptext);
	TECut(helptext);
	/* Copy in the new help text. */
	TESetText(helpstring, strlen(helpstring), helptext);
	/* Change it on the screen. */
	GetDItem(helpwin, diHelpText, &itemtype, &itemhandle, &itemrect);
	TextSize(14);
	TEUpdate(&itemrect, helptext);
	TextSize(18);
}

/* Respond to an event occurring in the help dialog. */

hit_help_dialog(dialog, itemhit, evt)
DialogPtr dialog;
short itemhit;
EventRecord *evt;
{
	HelpNode *prevhelpnode = curhelpnode;
	
	switch (itemhit) {
		case diHelpText:
			/* (nothing to do in the text right now) */
			return FALSE;
		case diHelpTopics:
			curhelpnode = firsthelpnode;
			break;
		case diHelpHelp:
			curhelpnode = helphelpnode;
			break;
		case diHelpNext:
			curhelpnode = curhelpnode->next;
			break;
		case diHelpPrevious:
			curhelpnode = curhelpnode->prev;
			break;
		case diHelpBack:
			if (nodestackpos > 0) {
				curhelpnode = nodestack[--nodestackpos];
			} else {
				/* close the help window?? */
			}
			break;
	}
	if (prevhelpnode != curhelpnode) {
		set_help_content(curhelpnode);
		if (itemhit != diHelpBack) {
			if (nodestackpos >= NODESTACKSIZE) {
				int i;
				
				for (i = 1; i < NODESTACKSIZE; ++i) nodestack[i - 1] = nodestack[i];
				nodestackpos = NODESTACKSIZE - 1;
			}
			nodestack[nodestackpos++] = prevhelpnode;
		}
	}
	return TRUE;
}
