/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

#include "conq.h"

/* Tmp array for counting terrain. */

int *numcells;

int badcountryplaces = 0;

int mungedterrain = 0;

int tmpradius;

int curmindistance;
int curmaxdistance;

int mungterrain;

int favorite[MAXUTYPES];

Unit *
find_place_for_occupant(side, u, x0, y0)
Side *side;
int u, x0, y0;
{
    Unit *transport;

    /* (should cast about randomly first, then do full search) */
    for_all_side_units(side, transport) {
	if (is_unit(transport)
	    && inside_area(transport->x, transport->y)
	    && type_can_occupy(u, transport)) {
	    return transport;
	}
    }
    return NULL;
}

/* Set the people inside the country to be on the country's side,
   flip a coin along the edges, to make the border more interesting. */
 
set_people_on_side(x, y)
int x, y;
{
    if ((distance(tmpside->startx, tmpside->starty, x, y) <= tmpradius
         || flip_coin())
        && probability(t_country_people(terrain_at(x, y)))) {
	set_people_side_at(x, y, side_number(tmpside));
    }
}

int growth;

expand_country(x, y)
int x, y;
{
    int u, t = terrain_at(x, y);
    Unit *unit;

    if (people_side_at(x, y) == NOBODY) {
	if (probability(t_country_growth(t))) {
	    for_all_unit_types(u) {
	    	tmputype = u;
		if (u_unit_growth(u) > random(10000)
		    && probability(ut_favored(u, t))
		    && valid_unit_place(x, y)) {
		    if ((unit = create_unit(u)) != NULL) {
			set_unit_side(unit, tmpside);
			enter_hex(unit, x, y);
		    	break;
		    }
		}
		if (u_indep_growth(u) > random(10000)
		    && probability(ut_favored(u, t))
		    && valid_unit_place(x, y)) {
		    if ((unit = create_unit(u)) != NULL) {
			enter_hex(unit, x, y);
		    	break;
		    }
		}
	    }
	    if (probability(t_country_people(t))) {
		set_people_side_at(x, y, side_number(tmpside));
	    }
	    ++growth;
	}
    } else {
	if (probability(t_country_takeover(t))) {
	    if (probability(t_country_people(t))) {
		set_people_side_at(x, y, side_number(tmpside));
	    }
	    /* look at stack and grab units in it */
	    ++growth;
	}
    }
}

int sideprogress;
int sidedeltahalf;

/* Place all the units belonging to countries.  Once placed, set ownership
   appropriately. */

make_countries()
{
    int x0, y0, u, t, i, tot, num, x, y;
    int canbeinopen[MAXUTYPES];  /* true if type need not be occ at start */
    int numtodo[MAXUTYPES], totnumtodo, totnumdone;
    int numleft[MAXUTYPES], neutsleft[MAXUTYPES], totleft, numlisted, numfails;
    Unit *unit, *transport;
    Side *side;
    int numdone = 0;
    int strength;
    char tmpbuf2[BUFSIZE];
    int favor;
    int dopeoplesides = FALSE, dopeopleindeps = FALSE;

    /* Run this always, unless something important is missing. */
    if (!terrain_defined()) return;
    announce_lengthy_process("Making countries");
    /* Precompute some important info */
    numcells = (int *) xmalloc(MAXTTYPES * sizeof(int));
    tot = 0;
    for_all_unit_types(u) {
	canbeinopen[u] = FALSE;
	favorite[u] = NONTTYPE;
	favor = 0;
	for_all_terrain_types(t) {
	    if (ut_favored(u, t) > 0) canbeinopen[u] = TRUE;
	    if (ut_favored(u, t) > favor) {
		favorite[u] = t;
		favor = ut_favored(u, t);
	    }
	}
	if (canbeinopen[u]) {
	    tot += u_start_with(u) + u_indep_near_start(u);
	}
    }
    /* Make space for people sides if we're going to have any. */
    if (!people_sides_defined()) {
	for_all_terrain_types(t) {
	    if (t_country_people(t) > 0) dopeoplesides = TRUE;
	    if (t_indep_people(t) > 0) dopeopleindeps = TRUE;
	}
	if (dopeoplesides || dopeopleindeps) {
	    allocate_area_people_sides();
	}
    }
    /* If no radius specified, pick something plausible. */
    if (g_min_radius() <= 0) {
	tmpradius = max(1, isqrt(3 * tot) / 2);
    } else {
	tmpradius = g_min_radius();
    }
    curmindistance = g_min_separation();
    curmaxdistance = g_max_separation();
    badcountryplaces = 0;
    mungedterrain = 0;
    for_all_sides(side) {
	sideprogress = (100 * numdone++) / numsides;
	sidedeltahalf = (100 / numsides) / 2;
	announce_progress(sideprogress);
	if (!country_is_complete(side)) {
	    strength = (side->player ? side->player->strength : 1);
	    /* Discover or generate the country's center. */
	    mungterrain = FALSE;
	    find_a_place(side, &x0, &y0);
	    announce_progress(sideprogress + sidedeltahalf);
	    Dprintf("%s country is at %d,%d\n", side_desig(side), x0, y0);
	    side->startx = x0;  side->starty = y0;
	    totnumtodo = totnumdone = 0;
	    for_all_unit_types(u) {
		numtodo[u] = 0;
		if (type_allowed_on_side(u, side)) {
		    numtodo[u] = u_start_with(u) * strength;
		    totnumtodo += numtodo[u];
		}
	    }
	    /* First do units actually belonging to the side initially. */
	    for_all_unit_types(u) {
		if (canbeinopen[u]) {
		    num = numtodo[u];
		    for (i = 0; i < num; ++i) {
			if (place_unit_type(u, x0, y0, &x, &y)) {
			    if ((unit = create_unit(u)) != NULL) {
				set_unit_side(unit, side);
				enter_hex(unit, x, y);
	   			announce_progress(sideprogress
						  + sidedeltahalf
						  + (sidedeltahalf
						     * totnumdone)
						  / totnumtodo);
			    }
			} else {
			    break;
			}
		    }
		    numleft[u] = num - i;
		} else {
		    numleft[u] = 0;
		}
	    }
	    /* Now do independents in the initial country area. */
	    for_all_unit_types(u) {
		if (type_allowed_on_side(u, NULL)) {
		    num = u_indep_near_start(u) * strength;
		    for (i = 0; i < num; ++i) {
			if (place_unit_type(u, x0, y0, &x, &y)) {
			    if ((unit = create_unit(u)) != NULL) {
				enter_hex(unit, x, y);
			    }
			} else {
			    break;
			}
		    }
		    neutsleft[u] = num - i;
		} else {
		    neutsleft[u] = 0;
		}
	    }
	    /* Now do units that have to be occupants. */
	    for_all_unit_types(u) {
		if (type_allowed_on_side(u, side) && !canbeinopen[u]) {
		    num = u_start_with(u) * strength;
		    for (i = 0; i < num; ++i) {
			if ((transport = find_place_for_occupant(side, u, x0, y0)) != NULL) {
			    if ((unit = create_unit(u)) != NULL) {
				set_unit_side(unit, side);
				enter_transport(unit, transport);
			    }
			} else {
			    break;
			}
		    }
		    numleft[u] = num - i;
		} else {
		    numleft[u] = 0;
		}
	    }
	    /* Now warn about what couldn't be placed. */
	    totleft = numlisted = numfails = 0;
	    sprintf(tmpbuf2, "");
	    for_all_unit_types(u) {
		totleft += numleft[u];
		if (numleft[u] > 0 && numlisted < 5) {
		    ++numfails;
		    strcat(tmpbuf2, " ");
		    strcat(tmpbuf2, u_type_name(u));
		}
	    }
	    if (totleft > 0) {
		char tmpbuf3[BUFSIZE];
		char *shortest_side_title();

		init_warning("could not put %d units in %s country (%s%s)",
			     totleft, shortest_side_title(side, tmpbuf3),
			     tmpbuf2, (numfails >= 5 ? " etc" : ""));
	    }
	}
	/* Now set the side of the people in this country. */
	tmpside = side;
	apply_to_area(side->startx, side->starty, tmpradius + 1,
		      set_people_on_side);
	/* Make sure each unit has people there, if allowed at all. */
	for_all_side_units(side, unit) {
	    if (inside_area(unit->x, unit->y)
	        && t_country_people(terrain_at(unit->x, unit->y)) > 0) {
		set_people_side_at(unit->x, unit->y, side_number(side));
	    }
	}
    }
    finish_lengthy_process();
    announce_lengthy_process("Growing countries");
    /* Grow each country out to a maximum radius. */
    for (i = tmpradius; i < g_radius_max(); ++i) {
    	announce_progress((100 * (i - tmpradius)) / (g_radius_max() - tmpradius));
	for_all_sides(side) {
	    tmpside = side;
	    growth = 0;
	    apply_to_ring(side->startx, side->starty, i-2, i, expand_country);
	    if (growth == 0 && probability(g_growth_stop())) {
	    	/* force this side to stop growing */
	    }
	}
    }
    finish_lengthy_process();
    /* Warn about any difficulties encountered. */
    if (badcountryplaces > 0) {
    	init_warning("%d of %d countries have undesirable locations",
		     badcountryplaces, numsides);
    }
    Dprintf("%d cells munged during placement\n", mungedterrain);
}

/* Test to see whether the side's pre-existing setup already suffices
   as a country in this game. */

country_is_complete(side)
Side *side;
{
    int u, numunits[MAXUTYPES];
    Unit *unit;
    int strength = (side->player ? side->player->strength : 1);

    for_all_unit_types(u) numunits[u] = 0;
    /* Should count only complete units or what? */
    for_all_side_units(side, unit) ++numunits[unit->type];
    for_all_unit_types(u) {
	if (numunits[u] < u_start_with(u) * strength) return FALSE;
    }
    return TRUE;
}

/* Count the hex as being of a particular type. */

count_hexes(x, y)
int x, y;
{
    ++numcells[terrain_at(x, y)];
}

/* Test whether a given location is desirable for a country.  It should be
   in the right distance range from other countries, and have enough of
   the right sorts of terrain. */

good_place(cx, cy)
int cx, cy;
{
    int toofar = TRUE, notfirst = FALSE;
    int c, px, py, u, t;
    Side *side;

    /* Check the candidate position against the other countries' positions. */
    for_all_sides(side) {
	px = side->startx;  py = side->starty;
	if (px > 0 && py > 0) {
	    notfirst = TRUE;
	    if (distance(cx, cy, px, py) < curmindistance) return FALSE;
	    if (distance(cx, cy, px, py) < curmaxdistance) toofar = FALSE;
	}
    }
    if (toofar && notfirst) return FALSE;
    /* Count the types of terrain in the country. */
    for_all_terrain_types(t) numcells[t] = 0;
    apply_to_area(cx, cy, tmpradius, count_hexes);
    /* Check against our upper and lower limits on terrain. */
    for_all_terrain_types(t) {
    	if (numcells[t] < t_country_min(t)) return FALSE;
	if (t_country_max(t) >= 0
	    && numcells[t] > t_country_max(t)) return FALSE;
    }
    return TRUE;
}

/* Work hard to find a place for a side's country.  First make some random
   trials, then start searching from the "center" of the area outwards.
   Then just pick a place and don't worry about it. */

find_a_place(side, cxp, cyp)
Side *side;
int *cxp, *cyp;
{
    int loosenings, tries, x, y;

    /* (what if side has most of a country already?) */

    Dprintf("%s country ", side_desig(side));
    for (loosenings = 0; loosenings < 4; ++loosenings) {
	for (tries = 0; tries < 200; ++tries) {
	    announce_progress(sideprogress + ((sidedeltahalf / 2) * tries) / 200);
	    random_point(&x, &y);
	    if (good_place(x, y)) { 
		*cxp = x;  *cyp = y;
		Dprintf("placed on try %d\n", tries);
		return;
	    }
	}
	if (search_around(area.width / 2, area.height / 2,
			  area.width / 2 + area.height / 2,
			  good_place, cxp, cyp, 1)) {
	    /* c[xy]p already written */
	    Dprintf("placed after search\n");
	    return;
	}
	curmindistance /= 2;
    }
    /* This should be a place in the area no matter what. */
    random_point(cxp, cyp);
    Dprintf("placed randomly\n");
    ++badcountryplaces;
    mungterrain = TRUE;
    return;
}    

/* The basic conditions that *must* be met by an initial unit placement. */

valid_unit_place(x, y)
int x, y;
{
    Unit *unit;
    int t = terrain_at(x, y);

    if (!inside_area(x, y)) return FALSE;
    if (ut_favored(tmputype, t) > 0
        && !ut_vanishes_on(tmputype, t)
        && !ut_wrecks_on(tmputype, t)
        && type_can_occupy_cell(tmputype, x, y)) {
        return TRUE;
    } else {
#if 0
    	/* Can't stack in the cell, but might be able to go inside a unit. */
	if (unit_at(x, y) != NULL) {
	    for_all_stack(x, y, unit) {
	    }
	}
#endif
    }
    return FALSE;
}

/* Find a place somewhere in the designated area, following constraints
   on terrain and adjacency.  Returns success of placement. */

place_unit_type(u, cx, cy, xp, yp)
int u, cx, cy, *xp, *yp;
{
    int t, tries, x, y, chance, r;
    tmputype = u;
    for (tries = 0; tries < 500; ++tries) {
	if (random_point_near(cx, cy, tmpradius, &x, &y)
	    && valid_unit_place(x, y)) {
	    chance = 100;
	    if (mungterrain
		&& favorite[u] != NONTTYPE
		&& probability(ut_favored(u, favorite[u]))) {
	    	set_terrain_at(x, y, favorite[u]);
	    	++mungedterrain;
	    }
	    t = terrain_at(x, y);
	    if (probability(chance)) {
	    	*xp = x;  *yp = y;
		Dprintf("%s placed on try %d\n", u_type_name(u), tries);
		return TRUE;
	    }
	}
    }
    /* Random shots aren't working, switch to exhaustive search and don't
       be picky about preferences. */
    if (search_around(cx, cy, tmpradius, valid_unit_place, &x, &y, 1)) {
	*xp = x;  *yp = y;
	Dprintf("Unit placed after search\n");
	return TRUE;
    }
    return FALSE;
}

/* A method that scatters independent units over the world. */

make_independent_units()
{
    int u, t, x, y;
    int doindeptype[MAXUTYPES], doanytype = FALSE, doindeppeople = FALSE;
    Unit *unit;

    /* Can't do anything without some terrain. */
    if (!terrain_defined()) return;
    /* Decide which types will be put down. */
    for_all_unit_types(u) {
    	doindeptype[u] = FALSE;
    	for_all_terrain_types(t) {
    	    if (ut_indep_density(u, t) > 0 && !ut_vanishes_on(u, t)) {
    	    	doindeptype[u] = TRUE;
    	    	doanytype = TRUE;
    	    	break;
    	    }
    	}
    }
    /* Make space for people sides if we're going to have any. */
    if (!people_sides_defined()) {
	for_all_terrain_types(t) {
	    if (t_indep_people(t) > 0) doindeppeople = TRUE;
	}
	if (doindeppeople) {
	    allocate_area_people_sides();
	}
    }
    /* If no units or peoples to do, don't waste time going through the world. */
    if (!doanytype && !doindeppeople) return;
    announce_lengthy_process("Making independents");
    /* Now apply the process to each hex individually. */
    for_all_interior_hexes(x, y) {
    	/* Progress is approx proportional to "column" being worked on. */
    	if (y == 1 && x % 5 == 0) announce_progress((x * 100) / area.width);
	t = terrain_at(x, y);
	for_all_unit_types(u) {
	    if (doindeptype[u]) {
		if (random(10000) < ut_indep_density(u, t)) {
	    	    if (!ut_vanishes_on(u, t)
			&& type_can_occupy_cell(u, x, y)) {
			if ((unit = create_unit(u)) != NULL) {
			    enter_hex(unit, x, y);
		    	}
		    }
		}
	    }
	}
	if (doindeppeople
	    /* Don't put indeps on top of people already on a side. */
	    && people_side_at(x, y) == NOBODY
	    && probability(t_indep_people(t))) {
	    /* 0 represents indep people present */
	    set_people_side_at(x, y, 0);
	}
    }
    finish_lengthy_process();
}
