/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact.  */

/* Interface implementations for the X11 version of xconq. */

#include "conq.h"
#include "xconq.h"

char *bitmaplib = BITMAPLIB;

/* Pixels of emptiness to leave around text etc. */

int margin = 2;
int fudge = 5;

/* Dimensions of a (pair of) map control item. */

int conwid = 40;
int conhgt = 20;

/* Fixed dimension of a scroll bar. */

int sbarwid = 16;

int rootcursor;                /* true if using parent window's cursor */

/* All the kinds of events we're normally interested in. */

int everything = KeyPressMask|ButtonPressMask|ButtonReleaseMask|ExposureMask;

/* The following magical arrays set all the sizes at each mag. */

/* (need flags to say what can be done at each mag) */

/* (define the range of mags also) */

/* Set up the basic user interface for a side.  This is called from the
   kernel while doing the final assignment of players to sides. */

init_ui(side)
Side *side;
{
    if (side_wants_display(side)) {
        side->ui = (UI *) xmalloc(sizeof(UI));
	/* Clear everything all at once (should just rely on xmalloc?). */
	bzero(side->ui, sizeof(UI));
	/* (should test ability to open display before continuing on) */
    } else {
	side->ui = NULL;
    }
}

/* Set up all sides' displays all at once. */

init_all_displays()
{
    Side *side;

    for_all_sides(side) {
	if (side_has_display(side)) {
	    init_display(side);
	}
    }
    Dprintf("One UI is %d bytes.\n", sizeof(UI));
}

/* Do a full redraw on each display. */

init_redraws()
{
    Side *side;

    for_all_sides(side) {
	if (side->ui != NULL) {
	    /* The moment of truth - up to now output has been suppressed. */
	    side->ui->active = TRUE;
	    map_all_windows(side);
	}
    }
}

/* Given a magnification power, update misc map parms from arrays. */

set_map_power(map, power)
Map *map;
int power;
{
    int upower = power;

    map->power = power;
    map->mag = mags[power];
    map->hw = hws[power];  map->hh = hhs[power];
    map->hch = hcs[power];
    map->uw = uws[upower];  map->uh = uhs[upower];
}

/* Open display, create all the windows we'll need, do misc setup things,
   and initialize some globals to out-of-range values for recognition later. */

/* (should ask what to do if display opens fail) */

int abort();

init_display(side)
Side *side;
{
    int i, p, u, t, s, n;
    char *name, *dpyname;
    Map *map;
    extern enum whattouse terrstyles[];
    Display *dpy;

    dpyname = side->player->displayname;

    Dprintf("Will try to open %s display \"%s\"...\n",
	    side_desig(side), dpyname);

    /* A hack. */
    if (strcmp("here", dpyname) == 0) {
	dpyname = getenv("DISPLAY");
    }
    /* Try to open the X display. */
    dpy = XOpenDisplay(dpyname);
    side->ui->dpy = dpy;
    /* If it failed, we won't get all the right participants in, so exit. */
    if (dpy == NULL) {
	/* We don't need to do a graceful shutdown here. */
	init_error("display \"%s\" could not be opened", dpyname);
	/* automatically exits */
    }
    if (DebugG) {
	XSynchronize(dpy, TRUE);
	XSetErrorHandler(abort);
	DGprintf("Synching the X server ...\n");
    }
    side->ui->screen = DefaultScreen(dpy);
    side->ui->rootwin = DefaultRootWindow(dpy);
    for (p = 0; p < NUMPOWERS; ++p) {
	for_all_terrain_types(t) {
	    /* Decide what display technique to use at each power. */
	    side->ui->usewhat[p][t] = terrstyles[p];
	    side->ui->terrpics[p][t] = None;
	    side->ui->terrchars[p][t] = '\0';
	    side->ui->terrfonts[p][t] = NULL;
	}
	for_all_unit_types(u) {
	    side->ui->unitpics[p][u] = None;
	    side->ui->unitmasks[p][u] = None;
	    side->ui->unitchars[p][u] = '\0';
	    side->ui->unitfonts[p][u] = NULL;
	    side->ui->ulegendfonts[p][u] = NULL;
	    /* (these values really should be filled in better) */
	    side->ui->unitw[p][u] = 1;  side->ui->unith[p][u] = 1;
	}
    }
    for (s = 0; s < MAXSIDES; ++s) {
	side->ui->emblempics[s] = None;
	side->ui->emblemmasks[s] = None;
    }
    side->ui->numlowcontrast = 0;
    side->ui->bonw = BLACKONWHITE;
#if 0
    /* get the name from the display unless the game is being restored from a 
       save file. (?) */
    if (!midturnrestore &&
	(name = XGetDefault(dpy, PROGRAMNAME, "SideName")) != NULL)
      side->name = name;
#endif
    /* Set up things shared by all the windows. */
    DGprintf("%d different colors available ...\n",
	     XDisplayCells(dpy, side->ui->screen));
    side->ui->monochrome = XDisplayCells(dpy, side->ui->screen) <= 2;
    set_colors(side);
    init_fonts(side);
    init_unit_images(side);
    init_terrain_images(side);
    init_emblem_images(side);
    init_bitmaps(side);
    init_gcs(side);
    init_cursors(side);
    /* Create the generic windows. */
    side->ui->maps = NULL;
    side->ui->front = NULL;
    /* (exact window config should come from resources/prefs) */
    create_map(side, 5, NULL /* areamap */);
#if 0
    /* Maybe open up a overall view of the area. */
    if (max(area.width, area.height) < 500 /* dpywid */) {
	/* We make a guess at an appropriate magnification,
	   based on the control panel size, which is fixed. */
	int power;
	int conhgt = 300; /* more or less */
	int tmp;

	tmp = conhgt / area.height;
	power = (tmp <= 1 ? 0 :
		 (tmp <= 2 ? 1 :
		  (tmp <= 4 ? 2 :
		   2 /* or 3? */))); 

	create_map(side, power, NULL);
    }
#endif
    /* Create a window where documentation-type info will go. */
    side->ui->help =
      XCreateSimpleWindow(dpy, side->ui->rootwin,
			  0, 0, 300, 600,
			  1, side->ui->fgcolor, side->ui->bgcolor);
    XSelectInput(dpy, side->ui->help, everything);
    XStoreName(dpy, side->ui->help, "Xconq Help");
    /* Flag some values as uninitialized */
    for (i = 0; i < MAXUTYPES; ++i) side->ui->bvec[i] = 0;
    Dprintf("Successfully inited \"%s\"!\n", dpyname);
}

/* Set up everybody's view of everybody else's colors and emblems. */

init_emblem_images(side)
Side *side;
{
    Side *side2;
    
    side->ui->eimages =
      (ImageFamily *) xmalloc(numsides * sizeof(ImageFamily));
    for_all_sides(side2) {
	init_emblem(side, side2);
    }
    /* Independent units may have a distinguishing emblem. */
    init_emblem(side, indepside);
}

/* This will set up the correct set of colors at any point in the game.
   Colors are all specified by name; if any are not available, it is up to
   the graphics interface to supply a substitute. */

set_colors(side)
Side *side;
{
    int	t;
    long fg, bg, tmp;
    char *colorname, substname[BUFSIZE];

    /* Our basic default is to have white on black. */
    fg = WhitePixel(side->ui->dpy, side->ui->screen);
    bg = BlackPixel(side->ui->dpy, side->ui->screen);
    if (0 /* side->ui->bonw */) {
	tmp = bg;  bg = fg;  fg = tmp;
    }
    side->ui->bgcolor = side->ui->owncolor = bg;
    side->ui->diffcolor = bg;
    side->ui->fgcolor = fg;
    side->ui->graycolor = side->ui->enemycolor = fg;
    side->ui->neutcolor = side->ui->goodcolor = side->ui->badcolor = fg;
    for_all_terrain_types(t) {
	if (side->ui->monochrome) {
	    side->ui->hexcolor[t] = side->ui->fgcolor;
	} else {
	    colorname = t_color(t);
	    /* Warn if we have to substitute a valid color name. */
	    if (colorname == NULL || strlen(colorname) == 0) {
		/* Substitute grays 33 to 99 only. */
		sprintf(substname, "gray%d", ((t * 66) / numttypes) + 33);
		init_warning("no color for terrain type \"%s\", using \"%s\"",
			     t_type_name(t), substname);
		colorname = substname;
	    }
	    side->ui->hexcolor[t] = request_color(side, colorname);
	}
    }
    /* Get special-purpose colors. */
    if (!side->ui->monochrome) {
	side->ui->owncolor = request_color(side, "black");
	side->ui->diffcolor = request_color(side, "maroon");
	side->ui->graycolor = request_color(side, "light gray");
	side->ui->enemycolor = request_color(side, "red");
	side->ui->neutcolor = request_color(side, "gray");
	side->ui->goodcolor = request_color(side, "green");
	side->ui->badcolor = request_color(side, "red");
    }
}

/* A predicate that tests whether our display can safely be written to. */

active_display(side)
Side *side;
{
    return (side && side->ui && side->ui->active);
}

/* Set up all the GCs. */

init_gcs(side)
Side *side;
{
    int u;
    unsigned int w, h;
    unsigned long mask;
    XGCValues values;
    GC gc;
    char *name;
    int junk1, junk2;
    Display *dpy = side->ui->dpy;
    Window rootwin = side->ui->rootwin;

    mask = GCForeground | GCBackground;
    values.foreground = side->ui->fgcolor;
    values.background = side->ui->bgcolor;

    gc = XCreateGC(dpy, rootwin, mask, &values);
    side->ui->gc = XCreateGC(dpy, rootwin, mask, &values);
    side->ui->textgc = XCreateGC(dpy, rootwin, mask, &values);
    side->ui->ltextgc = XCreateGC(dpy, rootwin, mask, &values);
    side->ui->helpgc = XCreateGC(dpy, rootwin, mask, &values);
    side->ui->icongc = XCreateGC(dpy, rootwin, mask, &values);
    side->ui->unitgc = XCreateGC(dpy, rootwin, mask, &values);
    /* for border drawing */
    values.line_width = 3;
    mask |= GCLineWidth;
    side->ui->varicongc = XCreateGC(dpy, rootwin, mask, &values);
    values.foreground = side->ui->bgcolor;
    values.background = side->ui->fgcolor;
    side->ui->invicongc = XCreateGC(dpy, rootwin, mask, &values);
    values.function = GXclear;
    mask = GCFunction;
    side->ui->cleargc = XCreateGC(dpy, rootwin, mask, &values);

    XSetFont(dpy, side->ui->gc, side->ui->textfont->fid);
    XSetFont(dpy, side->ui->textgc, side->ui->textfont->fid);
    XSetFont(dpy, side->ui->ltextgc, side->ui->textfont->fid);
    XSetFont(dpy, side->ui->helpgc, side->ui->helpfont->fid);
    if (side->ui->iconfont != NULL) {
	XSetFont(dpy, side->ui->icongc, side->ui->iconfont->fid);
    }
    DGprintf("done fonts ...\n");
    mask = GCFillStyle | GCGraphicsExposures;
    values.fill_style = FillSolid;
    values.graphics_exposures = FALSE;
    XChangeGC(dpy, side->ui->unitgc, mask, &values);
    XChangeGC(dpy, side->ui->varicongc, mask, &values);
    if (side->ui->unitfont != NULL) {
	XSetFont(dpy, side->ui->unitgc, side->ui->unitfont->fid);
    }
    if (side->ui->iconfont != NULL) {
	XSetFont(dpy, side->ui->invicongc, side->ui->iconfont->fid);
	XSetFont(dpy, side->ui->varicongc, side->ui->iconfont->fid);
    }

    /* Since clearbitgc has to have a depth of 1 we'll */
    /* hang it off bombpics[0] (is this still necessary??) */
    mask = GCFunction;
    values.function = GXclear;
    side->ui->clearbitgc =
      XCreateGC(dpy, side->ui->bombpics[0], mask, &values);

    XFreeGC(dpy, gc);
    DGprintf("GCs stored ...\n");
}

init_fonts(side)
Side *side;
{
    int p, u;
    XFontStruct *font;

    side->ui->textfont =
      open_font(side, TEXTFONT, "TextFont", NULL, "");
    side->ui->fw = font_width(side->ui->textfont);
    side->ui->fh = font_height(side->ui->textfont);
    side->ui->helpfont =
      open_font(side, HELPFONT, "HelpFont", side->ui->textfont, "text");
    side->ui->cursorfont =
      open_font(side, "cursor", "?", NULL, "");
    font = open_font(side, "*-bold-r-*-10-*", "LegendFont",
		     side->ui->textfont, "text");
	
    for (p = 0; p < NUMPOWERS; ++p) {
	for_all_unit_types(u) {
	    side->ui->ulegendfonts[p][u] = font;
	}
    }
    DGprintf("Fonts opened ...\n");
}

/* Open a font, possibly substituting another font if our desired one is
   missing for some reason. */

XFontStruct *
open_font(side, name, xdefault, altfont, alttype)
Side *side;
char *name, *xdefault, *alttype;
XFontStruct *altfont;
{
    char *altname;
    XFontStruct *font;

    /* A resource might override the given name. */
    /* Note that if override is bogus, then we won't go back to try the
       original name. */
    if ((altname = XGetDefault(side->ui->dpy, PROGRAMNAME, xdefault)) != NULL)
      name = altname;
    if ((font = XLoadQueryFont(side->ui->dpy, name)) == NULL) {
	init_warning("Can't open font \"%s\" on \"%s\"",
		     name, side->player->displayname);
	if (altfont != NULL) {
	    init_warning("Substituting the %s font, hope it works!", alttype);
	    return altfont;
	} else {
	    return NULL;
	}
    }
    DGprintf("Opened font \"%s\" ...\n", name);
    return font;
}

/* Get a total bounding box on the font. */

font_width(font)
XFontStruct *font;
{
    return font->max_bounds.width;
}

font_height(font)
XFontStruct *font;
{
    return font->max_bounds.ascent + font->max_bounds.descent;
}

string_width(str, font)
char *str;
XFontStruct *font;
{
    return (strlen(str) * font_width(font));
}

/* In X we need to "map" the windows so they'll actually be visible. */

map_all_windows(side)
Side *side;
{
    Map *map;

    for_all_maps(side, map) {
	XMapWindow(side->ui->dpy, map->window);
	XMapSubwindows(side->ui->dpy, map->window);
	/* Default mode is survey, set the cursor appropriately */
	if (!rootcursor) {
	    set_tool_cursor(side, map);
	}
    }
    /* Don't do side closeups. */
}

/* Get a color set up and warn if not getting what was asked for.  We can */
/* tolerate being off somewhat.  (Note X rgb value range is 0-65535.) */

#define CLOSE_ENOUGH(X,Y) (abs(((int) X) - ((int) Y)) < 2000)

long
request_color(side, name)
Side *side;
char *name;
{
    XColor c, avail;

    DGprintf("Requesting color %s\n", (name ? name : "<null>"));
    /* This might be called to get user-specified colors, even on a mono
       display, so deal with it. */
    if (name == NULL || strlen(name) == 0) {
	init_warning(
          "Requesting unnamed color on display \"%s\", substituting white",
		     side->player->displayname);
	return WhitePixel(side->ui->dpy, side->ui->screen);
    } else if (side->ui->monochrome) {
	if (strcmp("white", name) == 0) {
	    return WhitePixel(side->ui->dpy, side->ui->screen);
	} else if (strcmp("black", name) == 0) {
	    return BlackPixel(side->ui->dpy, side->ui->screen);
	} else {
	    init_warning(
              "No color \"%s\" on the mono display \"%s\", substituting white",
			 name, side->player->displayname);
	    return WhitePixel(side->ui->dpy, side->ui->screen);
	}
    } else {
	XAllocNamedColor(side->ui->dpy,
			 DefaultColormap(side->ui->dpy, side->ui->screen),
			 name, &avail, &c);
	if (!(CLOSE_ENOUGH(c.red, avail.red)
	      && CLOSE_ENOUGH(c.green, avail.green)
	      && CLOSE_ENOUGH(c.blue, avail.blue))) {
	    init_warning("%s color is way off on display \"%s\"!",
			 name, side->player->displayname);
	    init_warning("(%d %d %d instead of %d %d %d), but using it anyway",
			 avail.red, avail.green, avail.blue,
			 c.red, c.green, c.blue);
	}
	return avail.pixel;
    }
}

/* (flush for R5) */

/* Returns true if the two color values are hard to tell apart when they're
   adjacent on the screen. */

low_contrast(side, pv1, pv2)
Side *side;
unsigned long pv1, pv2;
{
    XColor c1, c2;

    if (pv1 == pv2) {
	return TRUE;
    } else if (side->ui->monochrome) {
	return FALSE;  /* is this really true? */
    } else if (side->ui->numlowcontrast == 0) {
	return FALSE;
    } else {
#if 0
	c1.pixel = pv1;
	XQueryColor(side->ui->dpy,
		    DefaultColormap(side->ui->dpy, side->ui->screen),
		    &c1);
	c2.pixel = pv2;
	XQueryColor(side->ui->dpy,
		    DefaultColormap(side->ui->dpy, side->ui->screen),
		    &c2);
	/* This is kind of heuristic... */
	if ((c1.red > 32000 && c2.red > 32000)
	    && (c1.green > 32000 && c2.green > 32000)
	    && (c1.blue > 32000 && c2.red > 32000))
	    return TRUE;
	if ((c1.red < 32000 && c2.red < 32000)
	    && (c1.green < 32000 && c2.green < 32000)
	    && (c1.blue < 32000 && c2.red < 32000))
	    return TRUE;
	/* Other tests should be made too. */
#endif
	return FALSE;
    }
}

/* Compute the distinctive emblem by which one side will recognize units
   of another side.  This does both our view of ourselves and others
   orthogonally.  Note that sides without displays can still have emblems
   and colors that the sides with displays will see, but that sides
   without displays don't need to do any emblem init. */

init_emblem(side, side2)
Side *side, *side2;
{
    int s2 = side_number(side2), i;
    ImageFamily *eimf;
    Image *eimg;

    /* Collect the color names of the other side and try to request
       them for our own display. */
    if (side2->colorscheme && strlen(side2->colorscheme) > 0) {
	side->ui->numcolors[2] = 1;
	/* eventually take spec apart and iterate for multiple colors */
	side->ui->colors[s2][0] = request_color(side, side2->colorscheme);
    }
    /* Turn emblem names into images. */
    if (side2->emblemname && strlen(side2->emblemname) > 0) {
	eimf = &(side->ui->eimages[s2]);
	eimf->name = side2->emblemname;
	load_image_family(side, eimf);
	/* Don't need a default emblem, drawing code will handle. */
	if (DebugG) {
	    describe_image_family(side, "emblem", side_desig(side2), eimf);
	}
    }
}

/* Shut the side's display down. */

close_display(side)
Side *side;
{
    XCloseDisplay(side->ui->dpy);
    side->ui->active = FALSE;
}

num_active_displays()
{
    int n = 0;
    Side *side;

    for_all_sides(side) {
	if (active_display(side)) ++n;
    }
    return n;
}
